import numpy as np
import xarray as xr
from pathlib import Path
from datetime import date, datetime, timedelta
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib import colors
from typing import List, Tuple

from gfatpy.lidar.nc_convert.measurement import merge_measurements_by_date
from gfatpy.lidar.utils.types import MeasurementType
from gfatpy.utils import plot
from gfatpy.utils.io import read_yaml_from_info
from gfatpy.utils.utils import parse_datetime
from gfatpy.lidar.utils.file_manager import filename2info
from gfatpy.lidar.utils.utils import LIDAR_INFO, signal_to_rcs
from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.plot.utils import get_norm, BoundsType


def quicklook_xarray(
    data_array: xr.DataArray,
    lidar_name: str,
    location: str = "Granada",
    is_rcs: bool = True,
    scale_bounds: BoundsType = "auto",
    colormap: str | colors.Colormap = "jet",
    ylims: Tuple[float, float] = (0.0, 14000.0),
    xlims: Tuple[datetime, datetime] | None = None,
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a lidar signal.

    Args:

        - data_array (xr.DataArray): Lidar signal or rcs from `lidar.preprocess`
        - is_rcs (bool, optional): To indicate if the data is RCS or not. Defaults to True.
        - scale_bounds (BoundsType, optional): scale bounds for the colorbar. Defaults to "auto".
        - colormap (str | matplotlib.colors.Colormap, optional): colormap of the colorbar. Defaults to "jet".

    Returns:

        - tuple[Figure, Axes]: Figure and Axes objects
    """

    # Convert cmap str to matplotlib colormap
    if isinstance(colormap, str):
        colormap = plt.get_cmap(colormap)
    cmap = colormap.copy()
    cmap.set_extremes(over="white")

    # Define channel
    channel = data_array.name.split("_")[1]

    try:
        if is_rcs:
            rcs = data_array.values
        else:
            rcs = signal_to_rcs(data_array.values, data_array.range.values)

        fig, ax = plt.subplots(figsize=(15, 5))
        norm, bounds = get_norm(
            rcs, scale_bounds, lidar_name=lidar_name, channel=channel
        )

        q = ax.pcolormesh(
            data_array.time,
            data_array.range / 1000.0,
            rcs.T,
            cmap=cmap,
            norm=norm,
        )

        ax.set_xlabel(r"Time, $[UTC]$")
        ax.set_ylabel(r"Height, $[km, \, agl]$")
        ax.set_ylim(ylims[0] / 1000.0, ylims[1] / 1000.0)

        # Set xlims as in the original xarray.Dataset dims (time)
        if xlims is not None:
            ax.set_xlim(xlims[0], xlims[1]) #type: ignore
        else:
            ax.set_xlim(data_array.time[0].values, data_array.time[-1].values)

        plot.watermark(ax, zoom=0.6, alpha=0.6)

        datestr = datetime.strftime(
            parse_datetime(data_array.time[1].values), "%Y-%m-%d"
        )
        ax.set_title(f"{location} | {lidar_name.upper()} | {datestr}")

        plot.apply_gap_size(ax, data_array=data_array)

        # Agregar la magnitud y unidades a la barra de color
        magnitud = f"RCS@{channel[:-3]}nm"
        unidades = "a.u."
        ticks = np.linspace(bounds.min(), bounds.max(), 9)
        colorbar = fig.colorbar(q, ticks=ticks, extend="max", pad=0.02)
        # fontsize colorbar
        colorbar.ax.tick_params(labelsize=12)

        # Crear el texto de magnitud y unidades
        text = f"{magnitud} [{unidades}]"

        # Calcular la mitad de la altura de la barra de color
        mid_height = (colorbar.ax.get_ylim()[0] + colorbar.ax.get_ylim()[1]) / 2

        # Añadir el texto a la barra de color girado 90 grados y 3 cm a la derecha de los valores
        x_position = 1.2  # Cambiar la coordenada x según sea necesario
        y_position = mid_height
        x_offset = 4  # 3 cm a la derecha (puede ajustarse)
        colorbar.ax.text(
            x_position + x_offset,
            y_position,
            text,
            rotation=90,
            va="center",
            ha="left",
            fontsize=12,
        )

        q.cmap.set_over("white")  # type: ignore
    except Exception as e:
        raise e
    return fig, ax


def quicklook_dataset(
    dataset: xr.Dataset,
    channel: str,
    /,
    scale_bounds: BoundsType = "auto",
    colormap: str | colors.Colormap = "jet",
    ylims: Tuple[float, float] = (0.0, 14000.0),
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a lidar signal.

    Args:

        - data_array (xr.DataArray): Lidar signal or rcs from `lidar.preprocess`
        - is_rcs (bool, optional): To indicate if the data is RCS or not. Defaults to True.
        - scale_bounds (BoundsType, optional): scale bounds for the colorbar. Defaults to "auto".
        - colormap (str | matplotlib.colors.Colormap, optional): colormap of the colorbar. Defaults to "jet".

    Returns:

        - tuple[Figure, Axes]: Figure and Axes objects
    """

    # Convert cmap str to matplotlib colormap
    if isinstance(colormap, str):
        colormap = plt.get_cmap(colormap)
    cmap = colormap.copy()
    cmap.set_extremes(over="white")

    data_array = dataset[f"signal_{channel}"]

    # Get lidar name from dataset
    lidar_name = dataset.attrs["location"].lower()

    quicklook_xarray(
        data_array,
        lidar_name=lidar_name,
        location="Granada",
        is_rcs=False,
        scale_bounds=scale_bounds,
        colormap=colormap,
        ylims=ylims,
    )

    try:
        rcs = signal_to_rcs(data_array.values, data_array.range.values)

        fig, ax = plt.subplots(figsize=(15, 5))
        norm, bounds = get_norm(
            rcs, scale_bounds, lidar_name=lidar_name, channel=channel
        )

        q = ax.pcolormesh(
            data_array.time,
            data_array.range / 1000.0,
            rcs.T,
            cmap=cmap,
            norm=norm,
        )

        ax.set_xlabel(r"Time, $[UTC]$")
        ax.set_ylabel(r"Height, $[km, \, agl]$")
        ax.set_ylim(ylims[0] / 1000.0, ylims[1] / 1000.0)

        datestr = datetime.strftime(parse_datetime(data_array.time[1].values), "%Y%m%d")
        plot.title1(f"{data_array.name} | {datestr}", 2)

        plot.watermark(ax, zoom=0.6, alpha=0.6)
        plot.apply_gap_size(ax, data_array=data_array)

        # Agregar la magnitud y unidades a la barra de color
        magnitud = "RCS"
        unidades = "a.u."
        ticks = np.linspace(bounds.min(), bounds.max(), 9)
        colorbar = fig.colorbar(q, ticks=ticks, extend="max")

        # Crear el texto de magnitud y unidades
        text = f"{magnitud} [{unidades}]"

        # Calcular la mitad de la altura de la barra de color
        mid_height = (colorbar.ax.get_ylim()[0] + colorbar.ax.get_ylim()[1]) / 2

        # Añadir el texto a la barra de color girado 90 grados y 3 cm a la derecha de los valores
        x_position = 1.2  # Cambiar la coordenada x según sea necesario
        y_position = mid_height
        x_offset = 4  # 3 cm a la derecha (puede ajustarse)
        colorbar.ax.text(
            x_position + x_offset,
            y_position,
            text,
            rotation=90,
            va="center",
            ha="left",
            fontsize=12,
        )

        q.cmap.set_over("white")  # type: ignore
    except Exception as e:
        raise e
    return fig, ax


def quicklook_from_file(
    filepath: Path | str,
    channels: List[str],
    output_dir: Path | str | None = None,
    scale_bounds: BoundsType = "auto",
    **kwargs,
):
    # Define output directory
    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)
    else:
        if not output_dir.exists():
            raise ValueError(f"Output directory {output_dir} does not exist.")

    # Check filepath
    if isinstance(filepath, str):
        filepath = Path(filepath)

    if not filepath.exists():
        raise FileNotFoundError(f"File {filepath} does not exist.")

    # Get lidar name from filepath
    nickname, _, _, _, _, target_date = filename2info(filepath.name)
    lidar_name = LIDAR_INFO["metadata"]["nick2name"][nickname]

    # Get INFO
    INFO = read_yaml_from_info(nickname, target_date)

    if kwargs.get("apply_ov", False):
        channels2load = channels.copy()
        for channel_ in channels2load:
            if "f" in channel_:
                overlap_channel_ = INFO["overlap_channels"][channel_]
                if overlap_channel_ not in channels2load:
                    channels2load.append(overlap_channel_)
    else:
        channels2load = None

    lidar = preprocess(
        filepath,
        channels=channels2load,
        apply_bg=kwargs.get("apply_bg", True),
        apply_dc=kwargs.get("apply_dc", False),
        apply_dt=kwargs.get("apply_dt", False),
        apply_bz=kwargs.get("apply_bz", True),
        apply_ov=kwargs.get("apply_ov", False),
    )

    if "ghk_dir" in kwargs and "depo_calib_dir" in kwargs:
        lidar.add_depolarization_products(
            lidar, ghk_dir=kwargs["ghk_dir"], depo_calib_dir=kwargs["depo_calib_dir"]
        )

    # Get location from dataset
    location = lidar.attrs["location"]

    # Get date string
    datestr = datetime.strftime(parse_datetime(lidar.time[1].values), "%Y%m%d_%H%M")

    for channel_ in channels:
        fig, _ = quicklook_xarray(
            lidar[f"signal_{channel_}"],
            lidar_name=lidar_name,
            location=location,
            is_rcs=False,
            scale_bounds=scale_bounds,
        )
        fig.savefig(
            output_dir / f"quicklook_{nickname}_{channel_}_{datestr}.png",
            dpi=600,
            bbox_inches="tight",
        )
        plt.close(fig)
    lidar.close()


def quicklook_from_date(
    lidar_name: str,
    channels: List[str],
    target_date: date | str,
    product_dir: Path | str,
    output_dir: Path | str | None = None,
    scale_bounds: BoundsType = "auto",
    **kwargs,
) -> List[Path]:

    if isinstance(target_date, str):
        target_date = parse_datetime(target_date).date()

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()
    output_dir.mkdir(parents=True, exist_ok=True)

    # Create xlim for target date from 0h to 0h of the next day
    xlims = (
        parse_datetime(target_date),
        parse_datetime(target_date + timedelta(days=1)),
    )

    lidar = merge_measurements_by_date(
        measurement_type=MeasurementType("RS"),
        lidar_name=lidar_name,
        target_date=target_date,
        product_dir=product_dir,
        channels=channels,
    )

    nickname = LIDAR_INFO["metadata"]["name2nick"][lidar_name]
    datestr = target_date.strftime("%Y%m%d")
    quicklook_paths = []
    for channel_ in channels:
        if channel_ not in lidar.channel.values:
            continue
        fig, _ = quicklook_xarray(
            lidar[f"signal_{channel_}"],
            lidar_name=lidar_name,
            is_rcs=False,
            scale_bounds=scale_bounds,
            ylims=kwargs.get("ylims", (0.0, 14000.0)),
            xlims=xlims,
        )
        quicklook_path = output_dir / f"quicklook_{nickname}_{channel_}_{datestr}.png"
        fig.savefig(
            quicklook_path,
            dpi=kwargs.get("dpi", 300),
            bbox_inches="tight",
        )
        plt.close(fig)
        quicklook_paths.append(quicklook_path)
    lidar.close()
    return quicklook_paths

def quicklook_lpdr(
    lpdr: xr.DataArray,    
    scale_bounds: BoundsType = (0, 0.5),
    colormap: str | colors.Colormap = "jet",
    ylims: Tuple[float, float] = (0.0, 15000.0),
    **kwargs,
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a lidar LPDR signal.

    Args:
        - lpdr (dict[int, xr.DataArray]): Dictionary containing LPDR data for different channels.
        - lidar_time (xr.DataArray): Time coordinate associated with the lidar data.
        - channel (int): Wavelength channel (e.g., 532 or 355).
        - scale_bounds (BoundsType, optional): Scale bounds for the colorbar. Defaults to (0, 0.5).
        - colormap (str | matplotlib.colors.Colormap, optional): Colormap for the colorbar. Defaults to "jet".
        - ylims (Tuple[float, float], optional): Y-axis limits in meters. Defaults to (0.0, 14000.0).

    Returns:
        - tuple[Figure, Axes]: Figure and Axes objects.
    """

    # Convert cmap str to matplotlib colormap
    if isinstance(colormap, str):
        colormap = plt.get_cmap(colormap)
    
    cmap = colormap.copy() #tye: ignore
    cmap.set_extremes(over="white")

    fig, ax = plt.subplots(figsize=(15, 5))
    norm, _ = get_norm(lpdr.values, scale_bounds)

    q = ax.pcolormesh(
        lpdr.time,  # Time coordinate
        lpdr.range / 1000.0,
        lpdr.values.T,  # Ensure correct shape for pcolormesh
        cmap=cmap,
        norm=norm,
        shading="auto"  # Allow automatic shading correction
    )

    ax.set_xlabel(r"Time, $[UTC]$")
    ax.set_ylabel(r"Range, $[km]$")
    ax.set_ylim(ylims[0] / 1000.0, ylims[1] / 1000.0)

    ticks = np.linspace(0, 0.5, 9)
    colorbar = fig.colorbar(q, ticks=ticks, extend="max")
    
    mid_height = (colorbar.ax.get_ylim()[0] + colorbar.ax.get_ylim()[1]) / 2
    x_position = 1.2  # Adjust as needed
    x_offset = 4
    colorbar.ax.text(
        x_position + x_offset, mid_height, "LPDR, [#]", rotation=90, va="center", ha="left", fontsize=12
    )

    q.cmap.set_over("white")  # type: ignore
    
    return fig, ax



import re
import sys
import shutil
import psutil
from pathlib import Path
from loguru import logger
from datetime import datetime

from gfatpy.lidar.utils.types import LidarName
from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.utils.io import find_nearest_filepath

RAW_FIRST_LETTER = LIDAR_INFO["metadata"]["licel_file_wildcard"]

logger.add(sys.stdout, level="INFO")


def search_config_file(
    lidar_name: LidarName,
    target_datetime: datetime,
    opt_config: Path | str | None,
) -> Path:
    """Searches for a configuration file.

    Args:
        lidar_name (LidarName): Name of the lidar.
        opt_config (Path | str | None, optional): Path to the configuration file. Defaults to None.

    Raises:
        FileNotFoundError: No configution file found in opt_config.

    Returns:
        Config: Configuration object.
    """
    if opt_config is not None:
        if isinstance(opt_config, Path):
            config_path = opt_config
        elif isinstance(opt_config, str):
            config_path = Path(opt_config)

        if not config_path.exists():
            raise FileNotFoundError(f"Configution file {opt_config} not found.")
    else:
        config_dir = Path(__file__).parent / "configs"
        config_path = find_nearest_filepath(
            config_dir,
            f"{lidar_name.value.upper()}*.toml",
            1,
            target_datetime,
            and_previous=True,
        )
        if not config_path.exists():
            raise FileNotFoundError(f"No configution file found in {config_dir}")
    return config_path


def cleanup_tmp_folders():
    """Delete temporary folders created during the process.

    Raises:

        - OSError: If there is an error deleting the temporary folder.
    """    
    pattern = r"tmp_unzipped_[a-zA-Z0-9_]+$"
    for dir_ in Path(__file__).parent.parent.parent.parent.glob("tmp_unzipped_*"):
        if re.search(pattern, dir_.name):
            try:
                # Ensure all files are closed
                for proc in psutil.process_iter(['pid', 'open_files']):
                    for file in proc.info['open_files'] or []:
                        if dir_ in Path(file.path).parents:
                            proc.terminate()
                            proc.wait()
                
                shutil.rmtree(dir_)
                print(f"Temporary folder deleted: {dir_}")
            except psutil.NoSuchProcess:
                print(f"Process already terminated.")
            except psutil.AccessDenied:
                print(f"Access denied to terminate process.")
            except RuntimeError as e:
                print(f"Error terminating process: {e}")


import os
import re
from tempfile import TemporaryDirectory
import numpy as np
from pathlib import Path
from datetime import datetime
from typing import overload, Any

import xarray as xr
from scipy.signal import savgol_filter
from scipy.integrate import cumulative_trapezoid as cumtrapz
from .types import LidarInfoType, Telescope

from gfatpy.utils.io import read_yaml

""" MODULE For General Lidar Utilities
"""

# LIDAR SYSTEM INFO
INFO_FILE = Path(__file__).parent.parent.absolute() / "info" / "info_lidars.yml"
LIDAR_INFO: LidarInfoType = read_yaml(INFO_FILE)

INFO_PLOT_FILE = Path(__file__).parent.parent.absolute() / "plot" / "info.yml"
LIDAR_PLOT_INFO = read_yaml(INFO_PLOT_FILE)


@overload
def signal_to_rcs(signal: xr.DataArray, ranges: xr.DataArray) -> xr.DataArray:
    ...


@overload
def signal_to_rcs(
    signal: np.ndarray[Any, np.dtype[np.float64]],
    ranges: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    ...


def signal_to_rcs(signal, ranges) -> xr.DataArray | np.ndarray[Any, np.dtype[np.float64]]:
    """Convert Lidar Signal to range-corrected signal

    Args:
        signal (np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray): Lidar signal
        ranges (np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray): Lidar ranges of signal

    Returns:
         xr.DataArray | np.ndarray[Any, np.dtype[np.float64]]: Range-corrected signal

    """
    return signal * ranges**2


@overload
def rcs_to_signal(rcs: xr.DataArray, ranges: xr.DataArray) -> xr.DataArray:
    ...


@overload
def rcs_to_signal(rcs: np.ndarray, ranges: np.ndarray) -> np.ndarray:
    ...


def rcs_to_signal(rcs, ranges):
    return rcs / ranges**2


def smooth_signal(signal, method="savgol", savgol_kwargs: dict | None = None):
    """Smooth Lidar Signal

    Args:
        signal ([type]): [description]
        method (str, optional): [description]. Defaults to 'savgol'.
    """

    if method == "savgol":
        if savgol_kwargs is None:
            savgol_kwargs = {"window_length": 21, "polyorder": 2}
        smoothed_signal = savgol_filter(signal, **savgol_kwargs)
    else:
        raise NotImplementedError(f"{method} has not been implemented yet")

    return smoothed_signal


def estimate_snr(signal, window=5):
    """[summary]

    Args:
        signal ([type]): [description]
    """

    # ventana: numero impar
    if window % 2 == 0:
        window += 1
    subw = window // 2

    n = len(signal)
    avg = np.zeros(n) * np.nan
    std = np.zeros(n) * np.nan

    for i in range(n):
        ind_delta_min = i - subw if i - subw >= 0 else 0
        ind_delta_max = i + subw if i + subw < n else n - 1

        si = signal[ind_delta_min : (ind_delta_max + 1)]
        avg[i] = np.nanmean(si)
        std[i] = np.nanstd(si)

        # print("%i, %i, %i" % (i, ind_delta_min, ind_delta_max + 1))
        # print(signal[ind_delta_min:(ind_delta_max+1)])
    snr = avg / std

    return snr, avg, std


def get_lidar_name_from_filename(fn):
    """Get Lidar System Name from L1a File Name
    Args:
        fn (function): [description]
    """
    lidar_nick = os.path.basename(fn).split("_")[0]
    if lidar_nick in LIDAR_INFO["metadata"]["nick2name"].keys():
        lidar = lidar_nick
    else:
        lidar = None
    return lidar


def to_licel_date_str(date: datetime) -> str:
    # Convert month number into hex capital letter
    month_hex = f"{date.month:x}".upper()
    return f'{date.strftime(r"%y")}{month_hex}{date.strftime(r"%d")}'


def licel_to_datetime(licel_name: str) -> datetime:
    """Convert Licel Date String to datetime
    Args:
        licel_name (str): Licel Date String
    """
    try:
        name, extension = licel_name.split(".")
    except ValueError:
        breakpoint()
    # Convert hexadecimal 'm' to decimal
    month_decimal = int(name[-5], 16)
    return datetime.strptime(
        f"{name[-7:-5]}{month_decimal:02d}{name[-4:-2]}T{name[-2:]}{extension[:4]}",
        r"%y%m%dT%H%M%S",
    )


def filter_wildcard(
    directory: Path | TemporaryDirectory, pattern_or_list: str | list[str] = r"\.\d+$", recursive_search=True
) -> list[Path]:
    """Filter files by wildcard

    Args:
        directory (Path): Directory to search.
        wildcard (str, optional): Wildcard. Defaults to r'\.\d+$'.
        recursive_search (bool, optional): Recursive search. Defaults to True means recursive search includes subdirectories.

    Raises:
        Exception: Directory does not exist.
        Exception: Directory is not a directory.

    Returns:
        list[Path]: List of files.
    """
    if isinstance(directory, TemporaryDirectory):
        directory = Path(directory.name)
        
    if not directory.exists():
        raise Exception(f"Directory {directory} does not exist")
    if not directory.is_dir():
        raise Exception(f"{directory} is not a directory")
    
    if isinstance(pattern_or_list, str):
        if recursive_search:            
            return [file for file in directory.rglob("*") if re.search(pattern_or_list, file.absolute().as_posix())]
        else:
            return [file for file in directory.glob("*") if re.search(pattern_or_list, file.absolute().as_posix())]
    else: 
        if recursive_search:
            return [file for file in directory.rglob("*") if file.name in pattern_or_list]
        else:
            return [file for file in directory.glob("*") if file.name in pattern_or_list]


def is_within_datetime_slice(filename: str, datetime_slice: slice) -> str:
    return datetime_slice.start <= licel_to_datetime(filename) < datetime_slice.stop

def get_532_from_telescope(telescope: Telescope = Telescope.xf) -> str:
    if telescope == telescope.xf:
        return "532xpa"
    elif telescope == telescope.ff:
        return "532fpa"
    elif telescope == telescope.nf:
        return "532npa"

    raise ValueError("Telescope type not recognized. Options are xf, ff, nf")


def sigmoid(x, x0, k, coeff: float = 1, offset: float = 0):
    y = 1 / (1 + np.exp(-k * (x - x0)))
    return (coeff * y) + offset


def extrapolate_beta_with_angstrom(
    beta_ref: np.ndarray,
    wavelength_ref: float,
    wavelength_target: float,
    angstrom_exponent: float | np.ndarray,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    return beta_ref * (wavelength_target / wavelength_ref) ** -angstrom_exponent


def integrate_from_reference(integrand, x, reference_index):
    """

    at x[ref_index], the integral equals = 0
    """
    # integrate above reference
    int_above_ref = cumtrapz(integrand[reference_index:], x=x[reference_index:])

    # integrate below reference
    int_below_ref = cumtrapz(
        integrand[: reference_index + 1][::-1], x=x[: reference_index + 1][::-1]
    )[::-1]

    return np.concatenate([int_below_ref, np.zeros(1), int_above_ref])


def optical_depth(extinction, height, ref_index=0):
    """
    Integrate extinction profile along height: r'$\tau(z) = \int_0^z d\dseta \alpha(\dseta)$'
    """

    return integrate_from_reference(extinction, height, reference_index=ref_index)


def refill_overlap(
    atmospheric_profile: np.ndarray[Any, np.dtype[np.float64]],
    height: np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray,
    fulloverlap_height: float = 600,
    fill_with: float | None = None,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Fill overlap region [0-`fulloverlap_height`] of the profile `atmospheric_profile` with the value `fill_with` provided by the user. If None, fill with the value at `fulloverlap_height`.

    Args:
        atmospheric_profile (np.ndarray): Atmospheric profile
        height (np.ndarray): Range profile in meters
        fulloverlap_height (float, optional): Fulloverlap height in meters. Defaults to 600 m.
        fill_with (float, optional): Value to fill the overlap region. Defaults to None.

    Returns:
        np.ndarray: Profile `atmospheric_profile` with the overlap region filled.
    """
    if isinstance(atmospheric_profile, xr.DataArray):
        atmospheric_profile = atmospheric_profile.values

    if fulloverlap_height < height[0] or fulloverlap_height > height[-1]:
        raise ValueError(
            "The fulloverlap_height is outside the range of height values."
        )

    idx_overlap = np.abs(height - fulloverlap_height).argmin()

    if fill_with is None:
        fill_with = atmospheric_profile[idx_overlap]

    new_profile = np.copy(atmospheric_profile)
    new_profile[:idx_overlap] = fill_with

    return new_profile



from pathlib import Path
import numpy as np
from matplotlib import colors
from loguru import logger
from typing import Literal

from gfatpy.utils.io import read_yaml

BoundsType = tuple[float, float] | Literal["auto", "limits", "from_info"]


def get_norm(
    rcs: np.ndarray,
    scale_bounds: BoundsType,
    color_resolution: int = 128,
    lidar_name: str | None = None,
    channel: str | None = None,
) -> tuple[colors.BoundaryNorm, np.ndarray]:
    """Get the color normalization and bounds for the colorbar.

    Args:

        - rcs (np.ndarray): Range corrected signal.
        - scale_bounds (BoundsType): Bounds for the colorbar.
        - color_resolution (int, optional): Colorbar resolution. Defaults to 128.

    Returns:

        - tuple[matplotlib.colors.BoundaryNorm, np.ndarray]: Color normalization and bounds.
    """
    match scale_bounds:
        case "auto":
            bounds = np.linspace(0, rcs.max() * 0.6, color_resolution)
        case "limits":
            bounds = np.linspace(rcs.min(), rcs.max(), color_resolution)
        case "from_info":
            if lidar_name is None or channel is None:
                raise ValueError("lidar_name and channel must be provided using option `from_info`.")
            PLOT_INFO = read_yaml(Path(__file__).parent / "info.yml")
            vmin = PLOT_INFO["limits"]["Vmin"][lidar_name][channel]
            vmax = PLOT_INFO["limits"]["Vmax"][lidar_name][channel]
            bounds = np.linspace(vmin, vmax, color_resolution)
        case _:
            bounds = np.linspace(*scale_bounds, color_resolution)

    norm = colors.BoundaryNorm(boundaries=bounds, ncolors=2**8, clip=False)
    logger.debug(f"Color bounds min - max: {bounds.min()} - {bounds.max()}")
    return norm, bounds



from datetime import datetime

import matplotlib.pyplot as plt
import numba
import numpy as np
import xarray as xr
from gfatpy.atmo.freudenthaler_molecular_properties import molecular_properties
from gfatpy.atmo.ecmwf import get_ecmwf_day
from gfatpy.utils.utils import adaptive_moving_average, parse_datetime
from typing import List
from scipy.signal import savgol_filter


def molecular_properties_2d(
    date: datetime | str,
    heights: np.ndarray,
    times: np.ndarray,
    wavelength: float = 532,
) -> xr.Dataset:
    """Calculate molecular properties for a given date, height ranges, and times.
    This function requests ECMWF (European Centre for Medium-Range Weather Forecasts) temperatures and pressures for a specified date and pipes them to `gfatpy.atmo.atmo.molecular_properties` to calculate molecular properties.

    Args:
        date (datetime | str): The date for which to request ECMWF data. Can be a datetime object or a string.
        heights (np.ndarray): An array of height ranges (in meters) for which to calculate molecular properties.
        times (np.ndarray): An array of times (in hours) for which to calculate molecular properties.
        wavelength (float, optional): The wavelength (in nanometers) for which to calculate molecular properties. Defaults to 532 nm.

    Returns:
        xr.Dataset: A dataset containing the following variables:
            - "molecular_beta": Molecular backscatter coefficient.
            - "molecular_alpha": Molecular extinction coefficient.
            - "attenuated_molecular_beta": Attenuated molecular backscatter coefficient.
            - "molecular_lidar_ratio": Molecular lidar ratio.
    """

    _date = parse_datetime(date)
    atmo_d = get_ecmwf_day(_date, heights=heights, times=times)

    mol_d = molecular_properties(
        wavelength,
        pressure=atmo_d.pressure.values,
        temperature=atmo_d.temperature.values,
        heights=atmo_d.range.values,
        times=atmo_d.time.values,
    )
    return mol_d


def iterative_fitting(
    rcs_profile: np.ndarray,
    attenuated_molecular_backscatter: np.ndarray,
    window_size_bins: int = 5,
    min_bin: int = 600,
    max_bin: int = 1000,
    debugger: bool = False,
) -> np.ndarray:
    """
    Perform iterative fitting on RCS and attenuated molecular backscatter profiles. 
    This function normalizes the RCS and attenuated molecular backscatter profiles
    and performs a linear fit over a sliding window. It calculates the slope of the
    fitted line for each window and returns a boolean matrix indicating the fit quality.
    
    Args:
    rcs_profile (np.ndarray): The range-corrected signal (RCS) profile.
    attenuated_molecular_backscatter (np.ndarray): The attenuated molecular backscatter profile.
    window_size_bins (int, optional): The size of the sliding window in bins. Default is 5.
    min_bin (int, optional): The minimum bin index to start the fitting. Default is 600.
    max_bin (int, optional): The maximum bin index to end the fitting. Default is 1000.
    debugger (bool, optional): If True, the function will plot the fitted lines and the original data. Default is False.

    Returns:
    np.ndarray: A boolean matrix indicating the fit quality.
    
    Raises:
    ValueError: If the shapes of `rcs_profile` and `attenuated_molecular_backscatter` do not match.
    """

    if rcs_profile.shape != attenuated_molecular_backscatter.shape:
        raise ValueError(f"RCS and Betta ranges must match")

    x_axis = np.arange(window_size_bins * 2)
    bool_matrix = np.full_like(rcs_profile, False, dtype=np.bool_)

    slope = []
    slope_mol = []

    for idx in np.arange(min_bin, max_bin):
        _rcs_norm = rcs_profile / rcs_profile[idx]
        _att_norm = (
            attenuated_molecular_backscatter / attenuated_molecular_backscatter[idx]
        )

        prof_slice = _rcs_norm[idx - window_size_bins : idx + window_size_bins]
        att_slice = _att_norm[idx - window_size_bins : idx + window_size_bins]

        coeff_prof = np.polyfit(np.arange(window_size_bins * 2), prof_slice, 1)
        coeff_att = np.polyfit(np.arange(window_size_bins * 2), att_slice, 1)

        slope.append(coeff_prof[0])
        slope_mol.append(coeff_att[0])


        att_data = np.polyval(coeff_att, x_axis)
        r2 = 1 - (
            ((prof_slice - att_data) ** 2).sum()
            / ((prof_slice - att_data.mean()) ** 2).sum()
        )

        if debugger:
            plt.scatter(x_axis, prof_slice, c="g")
            plt.plot(x_axis, np.polyval(coeff_prof, x_axis), c="g")
            plt.scatter(x_axis, att_slice, c="b")
            plt.plot(x_axis, np.polyval(coeff_att, x_axis), c="b")
            plt.show()
            print(f'Mol m: {coeff_att[0]}')
            print(f'Prof m: {coeff_prof[0]}')
            if r2 > 0.25:
                print(f"R^2: {r2}")

            plt.plot(slope)
            plt.plot(slope_mol)
            plt.close()
    return bool_matrix


def split_continous_measurements(
    time_array: np.ndarray, time_greater_than: float = 121 # Two minutest to avoid one profile search multiple DC
) -> list[np.ndarray]:
    
    """Groups times array into clusters with no more than `time_greater_than` seconds between measurements.
        
    Args:
        time_array (np.ndarray): Array of time measurements.
        time_greater_than (float, optional): Maximum allowed time difference between consecutive measurements to be considered in the same cluster. Defaults to 121 seconds.

    Returns:
        list[np.ndarray]: List of numpy arrays, each containing a cluster of time measurements.
    """
    diffs = (time_array[1:] - time_array[0:-1]).astype("f") / 1e9  # type: ignore
    return np.split(time_array, np.where(diffs > time_greater_than)[0] + 1)


def mask_by_slope(
    rcs: xr.DataArray,
    att_beta: xr.DataArray,
    min_height: float = 4000,
    max_height: float = 7000,
    window_size: float = 1000,
    window_time: float = 30,
    max_rel_error: float = 0.15,
    plot_profile: int | None = None,
) -> xr.DataArray:
    rcs_sel = rcs.sel(range=slice(min_height, max_height))
    beta_sel = att_beta.sel(range=slice(min_height, max_height))

    rcs_sel = xr.apply_ufunc(
        adaptive_moving_average, rcs_sel, kwargs={"window_size": window_time}, dask="allowed"
    )
    rcs_sel = xr.apply_ufunc(
        smooth_profiles, rcs_sel, kwargs={"window_size": 200}, dask="allowed"
    )

    result = np.full_like(beta_sel.values, False, dtype=bool)

    for idx, height in enumerate(rcs_sel.range):
        valid_ranges = rcs_sel.range[
            (height - window_size / 2 <= rcs_sel.range)
            & (height + window_size / 2 >= rcs_sel.range)
        ]
        height_rcs = rcs_sel.loc[dict(range=valid_ranges)]
        height_rcs /= height_rcs.isel(range=0)

        height_beta = beta_sel.loc[dict(range=valid_ranges)]
        height_beta /= height_beta.isel(range=0)

        x = np.arange(height_rcs.shape[1])

        slopes_rcs = np.polyfit(x, height_rcs.values.T, 1)[0]
        slopes_beta = np.polyfit(x, height_beta.values.T, 1)[0]

        result[:, idx] = (
            np.abs((slopes_rcs - slopes_beta) / slopes_beta) <= max_rel_error
        )

    result_data_array = xr.full_like(rcs, False)
    result_data_array.loc[dict(range=slice(min_height, max_height))] = result

    if plot_profile is not None:
        md = (max_height + min_height) / 2
        n_rcs = rcs[plot_profile] / rcs[plot_profile].sel(range=md, method="nearest")
        n_rcs_smth = rcs_sel[plot_profile] / rcs_sel[plot_profile].sel(
            range=md, method="nearest"
        )
        n_beta = att_beta[plot_profile] / att_beta[plot_profile].sel(
            range=md, method="nearest"
        )
        plt.plot(rcs.range, n_rcs)
        plt.plot(att_beta.range, n_beta)
        plt.plot(n_rcs_smth.range, n_rcs_smth)

        results = np.where(result_data_array[plot_profile], n_beta, np.nan)

        plt.plot(result_data_array.range, results, c="r", lw=2)

        plt.yscale("log")
        plt.show()

    return result_data_array


def mask_by_corrcoef(
    rcs: xr.DataArray,
    att_beta: xr.DataArray,
    min_height: float = 1000,
    max_height: float = 15000,
    window_size: float = 500,
    window_time: float = 15,
    min_corr: float = 0.95,
):
    """
    Masks a given range-corrected signal (RCS) DataArray based on the correlation coefficient 
    with an attenuated backscatter (att_beta) DataArray within a specified height range.
    
    Parameters:
    -----------
    rcs : xr.DataArray
        The range-corrected signal DataArray.
    att_beta : xr.DataArray
        The attenuated backscatter DataArray.
    min_height : float, optional
        The minimum height (in meters) for the selection range. Default is 1000.
    max_height : float, optional
        The maximum height (in meters) for the selection range. Default is 15000.
    window_size : float, optional
        The window size (in meters) for the correlation calculation. Default is 500.
    window_time : float, optional
        The window time (in minutes) for the adaptive moving average. Default is 15.
    min_corr : float, optional
        The minimum correlation coefficient threshold for masking. Default is 0.95.
    
    Returns:
    --------
    xr.DataArray
        A DataArray with the same shape as `rcs`, where values within the specified height 
        range are masked based on the correlation coefficient with `att_beta`.
    """
 
    rcs_sel = rcs.sel(range=slice(min_height, max_height))
    beta_sel = att_beta.sel(range=slice(min_height, max_height))

    rcs_sel = xr.apply_ufunc(smooth_profiles, rcs_sel, kwargs={"window_size": 170})
    rcs_sel = xr.apply_ufunc(
        adaptive_moving_average, rcs_sel, kwargs={"window_size": window_time}
    )

    ranges = rcs_sel.range.values
    
    range_mask_limits = [
        np.where((h - window_size / 2 <= ranges) & (h + window_size / 2 >= ranges))[0]
        for h in ranges
    ]

    lim_indexes = [
        [mask[0], mask[-1]] for mask in range_mask_limits
    ]

    result_array = windowed_correlation(
        rcs_sel.values,
        beta_sel.values,
        range_mask_limits=lim_indexes,
        min_corr=min_corr,
    )

    result_data_array = xr.full_like(rcs, False)
    result_data_array.loc[dict(range=slice(min_height, max_height))] = result_array

    return result_data_array
    # TODO:Finish function


def smooth_profiles(
    arr: np.ndarray, /, *, window_size: int = 11, polyorder: int = 3
) -> np.ndarray:
    """
    Smooths the profiles in a 2D array using the Savitzky-Golay filter.
    Parameters:
    arr (np.ndarray): A 2D array where each row represents a profile to be smoothed.
    window_size (int, optional): The length of the filter window (i.e., the number of coefficients). 
                                 Must be a positive odd integer. Default is 11.
    polyorder (int, optional): The order of the polynomial used to fit the samples. 
                               Must be less than window_size. Default is 3.
    Returns:
    np.ndarray: A 2D array with the same shape as `arr`, where each profile has been smoothed.
    """

    def smooth_profile(_x: np.ndarray) -> np.ndarray:
        return savgol_filter(_x, window_size, 3)

    return np.apply_along_axis(smooth_profile, 1, arr)


@numba.njit(parallel=True)
def windowed_correlation(
    rcs: np.ndarray,
    att_beta: np.ndarray,
    /,
    *,
    range_mask_limits: list[list[int]],
    min_corr: float,
) -> np.ndarray:
    """
    Compute the windowed correlation between two arrays with a given range mask and minimum correlation threshold.
    Parameters
    ----------
    rcs : np.ndarray
        The first input array for correlation computation.
    att_beta : np.ndarray
        The second input array for correlation computation.
    range_mask_limits : list[list[int]]
        A list of lists containing the start and end indices for masking the range of the input arrays.
    min_corr : float
        The minimum correlation threshold to determine if the correlation is significant.
    Returns
    -------
    np.ndarray
        An array of the same shape as `rcs` containing boolean values indicating whether the correlation
        at each point is greater than or equal to `min_corr`.
    """

    result_array = np.full(rcs.shape, np.nan)

    for t_idx in numba.prange(rcs.shape[0]):
        for r_idx in numba.prange(rcs.shape[1]):
            r_masks = range_mask_limits[r_idx]

            rcs_masked = rcs[t_idx][r_masks[0] : r_masks[1]]
            beta_masked = att_beta[t_idx][r_masks[0] : r_masks[1]]

            rcs_masked /= beta_masked[0]

            corr = np.corrcoef(rcs_masked, beta_masked)[0, 1]
            # set_trace()
            result_array[t_idx, r_idx] = corr >= min_corr

    return result_array


def cluster_value(arr: np.ndarray, /, *, value=1) -> list[list[tuple[int, int]]]:
    def cluster_1d(row) -> list[tuple[int, int]]:
        """
        Identifies clusters of consecutive elements in a 2D numpy array that match a specified value.
        Args:
            arr (np.ndarray): A 2D numpy array to be processed.
            value (int, optional): The value to identify clusters of. Defaults to 1.
        Returns:
            list[list[tuple[int, int]]]: A list of lists, where each inner list contains tuples representing
                                        the start position and length of each cluster found in the corresponding
                                        row of the input array.
        """
        _i: int = 0
        count: int = 0
        clusters: list[tuple[int, int]] = []  # list[position, count]
        for (_i, *_), _v in np.ndenumerate(row):
            if _v == value and count == 0:
                count += 1
            elif _v == value and count != 0:
                count += 1
            elif _v != value and count == 0:
                continue
            elif _v != value and count != 0:
                clusters.append((_i - (1 + count), count))
                count = 0

        if count != 0:
            clusters.append((_i - (1 + count), count))

        return clusters

    time_cluster: list[list[tuple[int, int]]] = []

    for time_row in arr:
        time_cluster.append(cluster_1d(time_row))

    return time_cluster


def cluster_at_least(
    clusters: list[list[tuple[int, int]]], n_min: int = 5, /
) -> np.ndarray:
    result = np.full((len(clusters), 2), np.nan, dtype=np.float64)
    """
    Filters and processes clusters based on a minimum threshold.
    This function takes a list of clusters, where each cluster is a list of tuples.
    Each tuple contains two integers. The function filters out tuples where the 
    second integer is less than `n_min`. For each cluster, it selects the tuple 
    with the highest second integer value that meets the threshold, and stores 
    the first integer and the sum of the first and second integers in the result 
    array.
    Parameters:
    -----------
    clusters : list[list[tuple[int, int]]]
        A list of clusters, where each cluster is a list of tuples containing two integers.
    n_min : int, optional
        The minimum threshold for the second integer in the tuples (default is 5).
    Returns:
    --------
    np.ndarray
        A 2D numpy array where each row corresponds to a cluster. Each row contains 
        two values: the first integer from the selected tuple and the sum of the 
        first and second integers from the selected tuple. If no tuple meets the 
        threshold in a cluster, the row contains NaN values.
    """

    for idx, prof_clust in enumerate(clusters):
        filtered = list(filter(lambda t: t[1] >= n_min, prof_clust))

        if len(filtered) == 0:
            continue

        selected = sorted(filtered, key=lambda t: t[1], reverse=True)[0]
        result[idx] = np.array([selected[0], selected[0] + selected[1]])

    return result




from ast import pattern
import xarray as xr
import re
import tempfile
from tkinter import W
import zipfile
import numpy as np
from pathlib import Path
from loguru import logger
from linc import get_config, write_nc_legacy
from linc.config.read import Config
from datetime import datetime as dt, time, timedelta, date


from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.utils.utils import (
    is_within_datetime_slice,
    licel_to_datetime,
    to_licel_date_str,
)
from gfatpy.utils.io import unzip_file
from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.lidar.utils.utils import filter_wildcard
from gfatpy.lidar.nc_convert.utils import search_config_file
from gfatpy.lidar.utils.file_manager import filename2info, info2general_path, info2path
from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope
from gfatpy.utils.utils import parse_datetime

RAW_FIRST_LETTER = LIDAR_INFO["metadata"]["licel_file_wildcard"]


class Measurement:
    path: Path
    type: MeasurementType
    lidar_name: LidarName
    telescope: Telescope
    _unzipped_dir: tempfile.TemporaryDirectory | None

    def __init__(
        self,
        path: Path,
        type: MeasurementType,
        lidar_name: str,
        telescope: str = "xf",
        **kwargs,
    ):
        self.path = path
        self.type = type
        self._is_zip = None
        self._filenames = None
        self.lidar_name = LidarName(lidar_name)
        self.telescope = Telescope(telescope)
        self._session_datetime = None
        self._datetimes = None
        self._unique_dates = None
        self._sub_dirs = None
        self._has_linked_dc = None
        self._dc = None
        self._filepaths = None
        self.config = None
        self._unzipped_dir = None
        self._is_linked = None

    @property
    def is_zip(self) -> bool:
        if self._is_zip is None:
            self._is_zip = self.path.suffix.endswith("zip")
        return self._is_zip

    @property
    def session_datetime(self) -> dt:
        if self._session_datetime is None:
            # Get datetime from filename RS_yyyymmdd_hhmm or RS_yyyymmdd_hhmm.zip
            self._session_datetime = dt.strptime(
                self.path.name.split(".")[0], f"{self.type.value}_%Y%m%d_%H%M"
            )
        return self._session_datetime

    @property
    def filenames(self) -> list[str]:
        """Get the filenames from the measurement directory (or zip file) withiout extracting them.

        Returns:
            list[str]: List of filenames.
        """

        if self.is_zip:
            with zipfile.ZipFile(self.path, "r") as zip_ref:
                self._filenames = sorted(
                    [
                        Path(file).name
                        for file in zip_ref.namelist()
                        if not (file.endswith("/") or file.endswith("dat"))
                    ]
                )
        else:
            self._filenames = sorted(
                [
                    file.name
                    for file in self.path.rglob("*.*")
                    if not (
                        file.name.endswith("/")
                        or file.name.endswith("dat")
                        or file.name.endswith("txt")
                    )
                ]
            )
        return self._filenames

    @property
    def sub_dirs(self) -> list[str]:
        """Extract sub-directories from the measurement path

        Returns:

            - list[str]: list of sub-directories.
        """
        if self._sub_dirs is None:
            folders = []
            path = self.path  # Store path in a local variable for use in the method
            if self.is_zip:
                with zipfile.ZipFile(path, "r") as zip_ref:
                    file_list = zip_ref.namelist()
                    folders = [
                        file.split("/")[-2] for file in file_list if file.endswith("/")
                    ]
            elif path.is_dir():
                folders = [f.name for f in path.iterdir() if f.is_dir()]
            self._sub_dirs = folders
        return self._sub_dirs

    @property
    def datetimes(self) -> list[dt]:
        """Get the dates from the measurement

        Returns:

            - list[dt]: List of dates
        """
        if self._datetimes is None:
            self._datetimes = sorted(
                [licel_to_datetime(filename_) for filename_ in self.filenames]
            )
        return self._datetimes

    @property
    def unique_dates(self) -> list[date]:
        """Get the unique dates from the measurement (not including the linked measurements)

        Returns:

            - list[dt]: List of unique dates
        """
        if self._unique_dates is None:
            self._unique_dates = np.unique(
                np.array([datetime_.date() for datetime_ in self.datetimes])
            ).tolist()
        return self._unique_dates

    @property
    def has_linked_dc(self) -> bool:
        """Check if the DC type measurement exists

        Returns:

            - bool: True if the DC type measurement exists. False otherwise.
        """
        if self._has_linked_dc is None:
            if self.type == MeasurementType.DC:
                self._has_linked_dc = False

            dc_path = self.path.parent / self.path.name.replace(
                self.type, MeasurementType.DC
            )
            self._has_linked_dc = dc_path.exists()
        return self._has_linked_dc

    @property
    def dc(self) -> "Measurement | None":
        """Get the DC type measurement

        Returns:

            - Measurement | None: DC type measurement
        """
        if self._dc is None:
            if self.type == MeasurementType.DC:
                self._dc = None
            else:
                if self.has_linked_dc:
                    dc_path = self.path.parent / self.path.name.replace(
                        self.type.value, MeasurementType.DC.value
                    )
                    self._dc = Measurement(
                        path=dc_path,
                        type=MeasurementType.DC,
                        lidar_name=self.lidar_name,
                    )
        return self._dc

    def get_filenames_within_datetime_slice(self, datetime_slice: slice) -> list[str]:
        return [
            filename
            for filename in self.filenames
            if is_within_datetime_slice(filename, datetime_slice)
        ]

    def unzip(
        self,
        pattern_or_list: str | list[str] = r"\.\d+$",
        destination: Path | None = None,
    ) -> Path | None:
        """Extract the zip file

        Args:

            - pattern_or_list (str, optional): pattern or list of patterns. Defaults to r'\\.\\d+$'.
            - destination (Path | None, optional): Directory to extract files. Defaults to None (extract to the same directory as the zip file).
        """
        if self.is_zip:
            if isinstance(pattern_or_list, str):
                _unzipped_dir = unzip_file(
                    self.path, pattern_or_list=pattern_or_list, destination=destination
                )
            else:
                _unzipped_dir = unzip_file(
                    self.path, pattern_or_list=pattern_or_list, destination=destination
                )                
        else:
            _unzipped_dir = None
        if isinstance(_unzipped_dir, tempfile.TemporaryDirectory):
            self._unzipped_dir = _unzipped_dir
            unzipped_dir = Path(_unzipped_dir.name)
        else:
            unzipped_dir = None
        return unzipped_dir

    def get_config(self, config_dir) -> Config:
        # Get lidar configuration file
        config_filepath = search_config_file(
            self.lidar_name, self.session_datetime, config_dir
        )
        return get_config(config_filepath)

    def has_target_date(self, target_date: dt | date | str) -> bool:
        """Check if the measurement has the target date

        Args:

            - date (dt): Target date

        Returns:

            - bool: True if the measurement has the target date. False otherwise.
        """
        if isinstance(target_date, str):
            date_ = parse_datetime(target_date).date()
        elif isinstance(target_date, dt):
            date_ = target_date.date()
        return date_ in self.unique_dates

    def generate_nc_output_path(
        self,
        output_dir: Path,
        target_date: dt | date | str | None = None,
        signal_type: MeasurementType | str = "rs",
        subdir: str = "",
        add_hour: bool = True,
    ) -> Path:
        """Generate the output dir for the netCDF file

        Args:

            - output_dir (Path): Directory to save the netCDF file.
            - signal_type (str): Signal type (rs :regular signal or sd: standard deviation).

        Returns:

            - Path: Output path for the netCDF file.
        """

        if isinstance(signal_type, str):
            signal_type = MeasurementType(signal_type.upper())

        if target_date is None:
            target_date = self.session_datetime
        elif isinstance(target_date, str):
            target_date = parse_datetime(target_date)
        elif isinstance(target_date, dt):
            target_date = target_date

        return info2path(
            lidar_name=self.lidar_name,
            date=self.session_datetime,
            measurement_type=self.type.value,
            signal_type=signal_type.value,
            telescope=self.telescope.value,
            dir=output_dir,
            subdir=subdir,
            add_hour=add_hour,
        )

    def _find_previous_paths(
        self, number_of_previous_days: int, lidar_name: LidarName, raw_dir: Path
    ) -> list[Path]:
        """Find the paths for the previous days

        Args:

             number_of_previous_days (int): Number of previous days to look for.
             lidar_name (LidarName): Lidar name (see gfatpy.lidar.nc_convert.types.LidarName).
             raw_dir (Path): Directory for the raw data.

        Returns:

             list[Path]: List of paths for the previous days.
        """
        current_date = self.session_datetime.date()
        prev_paths = [
            info2general_path(
                lidar_name.value,
                date=current_date - timedelta(days=n_day),
                data_dir=raw_dir,
            )
            for n_day in range(1, number_of_previous_days + 1)
        ]
        # Remove paths that do not exist
        previous_paths = [prev_path for prev_path in prev_paths if prev_path.exists()]
        return previous_paths

    def get_filepaths(
        self,
        pattern_or_list: str | list[str] = r"\.\d+$",
    ) -> set[Path] | None:
        """Get the filepaths from the measurement directory (or zip file) extracting them.

        Args:
            pattern_or_list (str, optional): Wildcard pattern or list of. Defaults to r"\\.\\d+$" (all). Example hour period (15:30-16:30) -> (r'RM\\d{6}(15[3-5]\\d|16[0-2]\\d)\\.\\d{7}')
            within_period (tuple[dt, dt] | None, optional): Initial (included) and final (excluded) hour of the period. Defaults to None (all files).
            include_linked_measurements (bool, optional): Include linked measurements. Defaults to True.

        Raises:
            Exception: If there are no files found in the measurement directory to meet the wildcard.

        Returns:
            set[Path]: Set of filepaths.
        """
        if self.is_zip:
            dir_ = self.unzip(pattern_or_list=pattern_or_list)
        else:
            dir_ = self.path

        if dir_ is None:
            raise Exception(
                f"No files found in {self.path} to meet the wildcard {pattern_or_list}"
            )

        if self.is_zip:
            found_files = set([*dir_.rglob("*.*")])
        else:
            found_files = set(filter_wildcard(dir_, pattern_or_list))
        return found_files

    def to_nc(
        self,
        pattern_or_list: str | list[str] = r"\.\d+$",
        output_dir: str | Path | None = None,
        config_dir: str | Path | None = None,
        filename: str | None = None,
        by_dates: bool = False,
        licel_pattern: str = "[A-Za-z]{2}\\d{2}[A-Za-z0-9]{1}\\d{4}\\.\\d{6}",
    ) -> list[Path] | None:
        """Convert the measurement to a netCDF file.

        Args:
            wildcard (str, optional): Wildcard of licel files (only for RS and HF measurements). Defaults to r"\\.\\d+$".
            output_dir (str | Path | None, optional): Output directory. Defaults to None (current working directory).
            config_dir (str | Path | None, optional): Configuration directory. Defaults to None (automatic search).
            include_linked_measurements (bool, optional): Include linked measurements. Defaults to True.
            filename (str | None, optional): Output filename. Defaults to None (automatic generation).
            by_dates (bool, optional): Split conversion by dates (only apply for RS and HF measurement types). Defaults to False.

        Raises:
            Exception: Error while converting the measurement.
            Exception: Error writing the netCDF file.
            RuntimeError: Error while converting the measurement.
            Exception: Error writing the netCDF file.

        Returns:
            list[Path] | None: List of output paths.
        """

        if isinstance(output_dir, str):
            output_dir = Path(output_dir)
        elif output_dir is None:
            output_dir = Path.cwd()

        # Get lidar configuration file
        config = self.get_config(config_dir=config_dir)

        if self.type == MeasurementType.RS or self.type == MeasurementType.HF:
            if by_dates:
                result_paths = []
                for date_ in self.unique_dates:
                    dstr = to_licel_date_str(dt.combine(date_, time(0, 0)))[:5]
                    pattern_: str = r"[A-Za-z]{2}" + f"{dstr}" + r"\d{2}\.\d{6}"
                    files2convert = self.get_filepaths(pattern_or_list=pattern_)
                    if files2convert is not None and len(files2convert) != 0:
                        files2convert = sorted(list(files2convert))
                        date_ = licel_to_datetime(
                            files2convert[0].name
                        )  # get date from first file
                        result_path = self.generate_nc_output_path(
                            target_date=date_,
                            output_dir=output_dir,
                            add_hour=True,
                        )

                        # Generate the output path
                        result_path.parent.mkdir(parents=True, exist_ok=True)
                        if filename is not None:
                            result_path = result_path.parent / filename

                        logger.info(f"Writing {result_path.name}")
                        # Write the nc file
                        try:
                            write_nc_legacy(
                                files2convert,
                                result_path,
                                config=config,
                            )
                        except Exception as e:
                            raise Exception(f"Error writing {result_path}: {e}")
                        result_paths.append(result_path)
            else:
                files2convert = self.get_filepaths(pattern_or_list=pattern_or_list)
                if files2convert is not None and len(files2convert) != 0:
                    files2convert = sorted(list(files2convert))
                    date_ = licel_to_datetime(
                        files2convert[0].name
                    )  # get date from first file
                    result_path = self.generate_nc_output_path(
                        target_date=date_,
                        output_dir=output_dir,
                    )
                    # Generate the output path
                    result_path.parent.mkdir(parents=True, exist_ok=True)
                    if filename is not None:
                        result_path = result_path.parent / filename

                    logger.info(f"Writing {result_path.name}")
                    # Write the nc file
                    try:
                        write_nc_legacy(
                            files2convert,
                            result_path,
                            config=config,
                        )
                    except Exception as e:
                        raise Exception(f"Error writing {result_path}: {e}")
                    result_paths = [result_path]
        elif self.type == MeasurementType.DC:
            files2convert = self.get_filepaths()
            if files2convert is not None and len(files2convert) != 0:
                result_path = self.generate_nc_output_path(
                    output_dir=output_dir,
                )
                result_path.parent.mkdir(parents=True, exist_ok=True)
                if filename is not None:
                    result_path = result_path.parent / filename
                logger.info(f"Writing {result_path.name}")

                # Write the nc file
                try:
                    write_nc_legacy(files2convert, result_path, config=config)
                except Exception as e:
                    raise Exception(f"Error writing {result_path}: {e}")
                result_paths = [result_path]
        else:
            result_paths = []
            for subdir in self.sub_dirs:
                if subdir == self.unique_dates[0].strftime("%Y%m%d"):
                    continue
                try:
                    pattern = (
                        r".*" + re.escape(f"{subdir}") + r".*" + licel_pattern
                    )
                    files2convert = self.get_filepaths(pattern_or_list=pattern)
                except:
                    raise RuntimeError(f"Error while converting: {self.path}")
                if files2convert is not None and len(files2convert) != 0:
                    result_path = self.generate_nc_output_path(
                        output_dir=output_dir,
                        subdir=subdir,
                    )
                    if filename is not None:
                        result_path = result_path.parent / filename
                    result_path.parent.mkdir(parents=True, exist_ok=True)
                    logger.info(f"Writing {result_path.name}")

                    # Write the nc file
                    try:
                        write_nc_legacy(files2convert, result_path, config=config)
                    except Exception as e:
                        raise Exception(f"Error writing {result_path}: {e}")
                    result_paths.append(result_path)
        return result_paths

    def remove_tmp_unzipped_dir(self):
        """Delete temporary folders created during the process.

        Raises:

            - OSError: If there is an error deleting the temporary folder.
        """
        if self._unzipped_dir is not None:
            try:
                to_be_deleted = Path(self._unzipped_dir.name)
                self._unzipped_dir.cleanup()
                self._unzipped_dir = None
                logger.info(f"Temporary folder deleted: {to_be_deleted}")
            except:
                logger.error(f"Error deleting temporary folder: {to_be_deleted.name}")

    # def __str__(self):
    #     return f"\nUnzipped Path: {self.unzipped_path}\n\n\n"
    def __str__(self):
        return f"Measurement Object\nPath: {self.path}\nType: {self.type}\nDates: {self.unique_dates}\nIs ZIP: {self.is_zip}\nHas DC: {self.has_linked_dc}\n"


def merge_measurements_by_date(
    lidar_name: str,
    target_date: dt | date | str,
    product_dir: str | Path,
    measurement_type: MeasurementType | None = None,
    output_dir: str | Path = Path.cwd(),
    channels: list[str] | None = None,
    **kwargs,
) -> xr.Dataset | xr.DataArray:

    if isinstance(product_dir, str):
        product_dir = Path(product_dir)
        logger.info(f"Searching measurements in {product_dir}")

    if isinstance(target_date, str):
        target_date = parse_datetime(target_date).date()
    elif isinstance(target_date, dt):
        target_date = target_date.date()
    logger.info(f"Target date: {target_date}")

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
        logger.info(f"Output directory: {output_dir}")

    if isinstance(measurement_type, str):
        measurement_type = MeasurementType(measurement_type)
    if measurement_type is None:
        raise ValueError("Measurement type is required.")

    # Create target date path
    path_with_measurements = []
    target_path = info2general_path(lidar_name, target_date, product_dir, data_level="1a")
    if not target_path.exists():
        logger.warning(
            f"No measurements found in {product_dir} for {lidar_name} on {target_date}"
        )
    else:
        path_with_measurements.append(target_path)

    # Create previous date
    prev_day = target_date - timedelta(days=1)
    prev_path = info2general_path(lidar_name, prev_day, product_dir, data_level="1a")

    if prev_path.exists():
        path_with_measurements.append(prev_path)

    if path_with_measurements == []:
        raise FileNotFoundError(
            f"No measurements found in {product_dir} for {lidar_name} on {target_date}"
        )

    logger.info(f"Searching measurements in previous target date path: {prev_path}")
    result_paths = []
    for path_ in path_with_measurements:
        for m_path in path_.glob("*"):
            result_paths.append(m_path)

    measures = []
    for m_path in result_paths:
        if m_path is None:
            continue
        _, _, type_, _, _, _ = filename2info(m_path.name)
        if measurement_type == MeasurementType(type_.upper()):
            measures.append(m_path)

    lidars = []
    crop_ranges = kwargs.get("crop_ranges", (0, 14000.0))
    for file in measures:
        lidar_ = preprocess(
            file, channels=channels, crop_ranges=crop_ranges, gluing_products=False
        )
        lidars.append(lidar_)

    merge_dataset = xr.combine_by_coords(
        [lidar_ for lidar_ in lidars], combine_attrs="drop_conflicts"
    )
    return merge_dataset


def info2measurements(
    lidar_name: str,
    target_date: dt | date | str,
    raw_dir: str | Path,
    measurement_type: MeasurementType | None = None,
) -> list[Measurement] | None:
    """Converts a list of paths to a list of Measurement objects.

    Args:
        lidar_name (str): _description_
        target_date (dt | date | str): _description_
        raw_dir (str | Path): _description_
        measurement_type (MeasurementType | None, optional): _description_

    Returns:
        list[Measurement] | None: _description_
    """

    if isinstance(raw_dir, str):
        raw_dir = Path(raw_dir)
        logger.info(f"Searching measurements in {raw_dir}")

    if isinstance(target_date, str):
        target_date = parse_datetime(target_date).date()
    elif isinstance(target_date, dt):
        target_date = target_date.date()
    logger.info(f"Target date: {target_date}")

    if isinstance(measurement_type, str):
        measurement_type = MeasurementType(measurement_type)

    list_paths = set()

    # Create target date path
    target_path = info2general_path(lidar_name, target_date, raw_dir)
    if target_path.exists():
        logger.info(f"Searching measurements in target path: {target_path}")
        list_paths.update([*target_path.glob("*")])

    # Looking for previous days
    prev_day = target_date - timedelta(days=1)
    prev_path = info2general_path(lidar_name, prev_day, raw_dir)

    if prev_path.exists():
        logger.info(f"Searching measurements in previous target date path: {prev_path}")
        list_paths.update([*prev_path.glob("*")])

    if list_paths == []:
        logger.warning(
            f"No measurements found in {raw_dir} for {lidar_name} on {target_date}"
        )
        return None

    logger.info(f"Number of measurement candidates: {len(list_paths)}")
    # Filter by type
    if measurement_type is not None:
        list_paths = [
            path_ for path_ in list_paths if measurement_type.value in path_.name
        ]

    logger.info(
        f"Number of measurement candidates after filtering by type: {len(list_paths)}"
    )

    measurements = []
    for path_ in list_paths:
        if len(path_.name.split(".")[0]) == 16:
            logger.info(f"Checking candidate measurement: {path_}")
            m_ = Measurement(
                path=path_,
                type=MeasurementType(path_.name[:2]),
                lidar_name=lidar_name,
            )
            if target_date in m_.unique_dates:
                logger.info(f"Candidate measurement accepted: {m_.path.name}")
                measurements.append(m_)
            else:
                logger.warning(f"Candidate measurement rejected: {path_.name}")
        else:
            logger.warning(f"Invalid measurement name (skipped): {path_.name}")
    if measurements == []:
        return None
    return measurements


def filter_by_type(
    measurements: list[Measurement], mtype: MeasurementType
) -> list[Measurement]:
    """Filter a list of measurements by type.

    Args:

        - measurements (list[Measurement]): List of measurements to filter.
        - mtype (MeasurementType): Type to filter by.

    Returns:

        - list[Measurement]: Filtered list of measurements.
    """
    return list(
        filter(
            lambda m: m.type == mtype,
            measurements,
        )
    )



import numpy as np
from numpy.polynomial.polynomial import polyfit

import numba

from gfatpy.utils.utils import linear_fit


def best_slope_fit(mat1: np.ndarray, mat2: np.ndarray, window: int) -> np.ndarray:
    """Best slope fit

    Args:
        mat1 (np.ndarray): Matrix 1
        mat2 (np.ndarray): Matrix 2
        window (int): Window size

    Returns:
        np.ndarray: indexes the best slope fit is found in axis=1. Results in axis=0.
    """

    assert mat1.shape == mat2.shape, "Matrices shape must match"
    assert isinstance(window, int), "Window argument must be an integer"

    x = np.arange(window)
    res = np.array([])

    # Iter in first dimension, equivalent to time in lidar case
    for idx in range(mat1.shape[0]):
        windowed1 = rolling(mat1[idx], window)
        windowed2 = rolling(mat2[idx], window)

        slopes1 = polyfit(x, windowed1.T, 1)[0]
        slopes2 = polyfit(x, windowed2.T, 1)[0]
        chosen_group = np.argmin(np.abs(slopes1 - slopes2) / slopes1)
        res = np.hstack([res, chosen_group + np.floor(window / 2)])  # Append chosen bin

    return res.astype(int)


@numba.njit(parallel=True)
def windowed_corrcoefs(arr1: np.ndarray, arr2: np.ndarray, w_size: int):
    """
    Compute windowed correlation coefficients between two 2D arrays using a sliding window approach.

    Parameters:
    arr1 (np.ndarray): First input array of shape (n_samples, n_features).
    arr2 (np.ndarray): Second input array of shape (n_samples, n_features).
    w_size (int): Size of the sliding window.

    Returns:
    np.ndarray: Array of shape (n_samples, n_windows) containing the correlation coefficients for each window.
    """
    range_shape = arr1.shape[1] - (w_size - 1)
    _corrcoefs = np.empty((arr1.shape[0], range_shape))
    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)
        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]
            coeff = np.corrcoef(_w1, _w2)[1, 0]
            _corrcoefs[t_idx][idx] = coeff
    return _corrcoefs


@numba.njit(parallel=True)

def windowed_proportional(arr1: np.ndarray, arr2: np.ndarray, /, *, w_size: int):
    """
    Compute the windowed proportional factors and their deviations between two matrices.
    This function calculates the proportional factors and their deviations for each window
    of size `w_size` in the input matrices `arr1` and `arr2`. The function uses Numba's 
    Just-In-Time (JIT) compilation to optimize performance and parallel execution.
    Parameters:
    -----------
    arr1 : np.ndarray
        The first input matrix. Must have the same shape as `arr2`.
    arr2 : np.ndarray
        The second input matrix. Must have the same shape as `arr1`.
    w_size : int
        The size of the window to use for the rolling calculations.
    Returns:
    --------
    _factor : np.ndarray
        A matrix containing the proportional factors for each window.
    _proportional : np.ndarray
        A matrix containing the mean absolute proportional deviations for each window.
    Raises:
    -------
    AssertionError
        If the shapes of `arr1` and `arr2` do not match.
    """
    assert arr1.shape == arr2.shape, "Matrices shape must match"
    # assert isinstance(w_size, int), "Window argument must be an integer"

    range_shape = arr1.shape[1] - (w_size - 1)
    _proportional = np.full((arr1.shape[0], range_shape), np.nan)
    _factor = np.full((arr1.shape[0], range_shape), np.nan)

    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)

        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]

            ratio = np.mean(_w2 / _w1)
            adj = _w1 * ratio
            _factor[t_idx][idx] = ratio
            _proportional[t_idx][idx] = (np.abs(adj - _w2) / _w2).mean()

    return _factor, _proportional


@numba.njit()
def rolling(a, window):
    """
    Apply a rolling window to a 1D numpy array.

    Parameters:
    a (numpy.ndarray): Input 1D array.
    window (int): Size of the rolling window.

    Returns:
    numpy.ndarray: A 2D array where each row is a windowed segment of the input array.

    Example:
    >>> import numpy as np
    >>> a = np.array([1, 2, 3, 4, 5])
    >>> rolling(a, 3)
    array([[1, 2, 3],
        [2, 3, 4],
        [3, 4, 5]])
    """
    shape = (a.size - window + 1, window)
    strides = (a.itemsize, a.itemsize)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


@numba.njit
def correlate_vector_to_matrix(vector, matrix):
    """Correlate a vector to a matrix

    Args:
        vector (np.ndarray): Vector to correlate
        matrix (np.ndarray): Matrix to correlate

    Returns:
        np.ndarray: Correlation result
    """
    corr = np.zeros(matrix.shape[1])
    for i in range(matrix.shape[1]):
        corr[i] = np.corrcoef(vector, matrix[:, i])[0, 1]
    return corr


def rolling_window_test(a, window):
    """
    Create a rolling window view of the input array.

    Parameters:
    a (numpy.ndarray): Input array.
    window (int): Size of the rolling window.

    Returns:
    numpy.ndarray: A view of the input array with the rolling window applied.

    Example:
    >>> import numpy as np
    >>> a = np.array([1, 2, 3, 4, 5])
    >>> rolling_window_test(a, 3)
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])
    """
    shp = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shp, strides=strides)

def moving_linear_fit(x_array, y_array, window_size, **kwargs):
    """
    Perform a moving linear fit on the given data arrays.
    Parameters:
    x_array (np.ndarray): The array of x-values.
    y_array (np.ndarray): The array of y-values.
    window_size (int): The size of the moving window.
    **kwargs: Additional keyword arguments.
    Returns:
    dict: A dictionary containing the following keys:
        - "slope" (np.ndarray): The slope of the linear fit for each window.
        - "std_slope" (float): The standard deviation of the slope.
        - "durbin_watson" (np.ndarray): The Durbin-Watson statistic for each window.
        - "mrse" (np.ndarray): The mean squared relative error for each window.
        - "anderson" (np.ndarray): The Anderson-Darling statistic for each window.
    """ 
    xdata = rolling_window_test(x_array, window_size).T
    ydata = rolling_window_test(y_array, window_size).T
    
    d, slope = np.nan*np.ones(len(x_array)), np.nan*np.ones(len(x_array))
    mrse, anderson_coef = np.nan*np.ones(len(x_array)), np.nan*np.ones(len(x_array))
    for idx in range(ydata.shape[1]):
        # fit_parameters = np.polyfit(xdata[:,idx], ydata[:,idx], deg = degree, full=False)
        stats = linear_fit(xdata[:,idx], ydata[:,idx])
        slope[idx]  = stats["parameters"][1] 
        std_slope = stats["standard_deviation_parameters"][1]
        d[idx] = stats["durbin_watson"]
        mrse[idx] = stats["msre"]
        anderson_coef[idx] = stats["anderson"][0]
        if kwargs.get("debugger", False):
            ranges = [1100, 1300, 1600, 1900]
            if idx in ranges:
                import matplotlib.pyplot as plt
                fig, ax = plt.subplots()
                ax.plot(xdata[:,idx], ydata[:,idx], linewidth=0, marker='o', label=f'{ranges[idx]}')
                ax.plot(xdata[:,idx],np.polyval(np.flip(stats["parameters"]), xdata[:,idx]), label='fit')
                fig.savefig(f'test_dws_{idx}.png', dpi=300)
                plt.close(fig)
    results = {"slope": slope, "std_slope": std_slope ,"durbin_watson": d, "mrse": mrse, "anderson": anderson_coef}
    return results



from typing import Callable
from warnings import warn


from loguru import logger
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter

from gfatpy.utils.optimized import best_slope_fit
from gfatpy.utils.utils import adaptive_moving_average


def create_smoother(window_size: int) -> Callable[[np.ndarray], np.ndarray]:
    def smooth(row: np.ndarray) -> np.ndarray:
        return savgol_filter(row / np.nanmean(row), window_size, 4)

    return smooth


def take_mid_value(arr: np.ndarray) -> float:
    shape = arr.shape[0]
    return arr[shape // 2]


def gluing(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    range_min: float = 1500,
    range_max: float = 5000,
    criteria_half_window: float = 1000,
    adjustment_half_window: float = 1000,
    shift_ratio: float = 0.4
):

    sel_signal_an = signal_an.sel(
        range=slice(range_min - criteria_half_window, range_max + criteria_half_window)
    )
    sel_signal_pc = signal_pc.sel(
        range=slice(range_min - criteria_half_window, range_max + criteria_half_window)
    )

    sel_range = sel_signal_pc.range.values
    window_size_bins = int((criteria_half_window * 2) / (sel_range[1] - sel_range[0]))

    sm_sel_signal_an = adaptive_moving_average(sel_signal_an.values, window_sizes=10.)
    sm_sel_signal_pc = adaptive_moving_average(sel_signal_pc.values, window_sizes=10.)
        
    signal_glued = signal_pc.values.copy()

    gluing_bins = best_slope_fit(
        sm_sel_signal_an / take_mid_value(sm_sel_signal_an),
        sm_sel_signal_pc / take_mid_value(sm_sel_signal_pc),
        window=window_size_bins,
    )

    gluing_ranges = sel_range[gluing_bins]
    
    gluing_ranges = np.full_like(gluing_ranges, fill_value=np.median(gluing_ranges))

    for idx, gluing_range in enumerate(gluing_ranges[:]):
        near_idx = (
            sel_range > (gluing_range - shift_ratio * adjustment_half_window)
        ) & (sel_range < (gluing_range + (1 + shift_ratio) * adjustment_half_window))

        # near_idx = (
        #     sel_range > (gluing_range - adjustment_half_window)
        # ) & (sel_range < (gluing_range + adjustment_half_window))
        
        # sel_an = sel_signal_an[dict(time=idx)][near_idx]
        # sel_pc = sel_signal_pc[dict(time=idx)][near_idx]

        sel_an = sm_sel_signal_an[idx, near_idx]
        sel_pc = sm_sel_signal_pc[idx, near_idx]

        r_gluing = np.corrcoef(sel_an, sel_pc)[0, 1]

        fit_values = np.polyfit(
            sel_an,  # type: ignore
            sel_pc,  # type: ignore
            1,
        )

        # plt.scatter(sel_an, sel_pc)
        # plt.plot(sel_an, np.polyval(fit_values, sel_an), label=f"{r_gluing}")
        # plt.legend()
        # plt.show()
        
        if r_gluing > 0.9:  # TODO: Adjust threshold
            fit_values[-1] = 0
            signal_an2pc = np.polyval(fit_values, signal_an.values[idx, :])
            signal_glued[idx, :] = np.concatenate(
                [
                    signal_an2pc[signal_pc.range.values < gluing_range],
                    signal_pc[idx, signal_pc.range.values >= gluing_range],
                ]
            )
        else:            
            breakpoint()
            logger.warning(
                f"r2 thereshold condition not satisfied for profile in time index {idx}"
            )
            # TODO: warn or implement filling method
    return signal_glued



import pathlib
from datetime import datetime
from typing import List, Optional, Tuple
from pathlib import Path

import typer

from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope
from gfatpy.lidar.reader import reader_xarray
from gfatpy.lidar.quality_assurance.rayleigh_fit import rayleigh_fit_from_file


app = typer.Typer(no_args_is_help=True)


@app.command(
    help="Converts raw lidar data to l1 data",
    no_args_is_help=True,
)

@app.command(help="QA Rayleigh fit", no_args_is_help=True)
def rayleigh_fit(
    file: Path = typer.Option(
        ..., "--file", "-f", readable=True, dir_okay=False, file_okay=True
    ),
    initial_hour: int = typer.Option(None, "--initial-hour", "-i"),
    duration: int = typer.Option(None, "--duration", "-d"),
    channels: Optional[List[str]] = typer.Option(None, "--channels", "-c"),
    reference_range: Optional[Tuple[float, float] | None] = typer.Option(
        None, "--reference-range", "-r"
    ),
    smooth_window: Optional[float | None] = typer.Option(None, "--smooth-window", "-s"),
    save_fig: bool = typer.Option(False, "--save-fig", "-g"),
    output_dir: Path = typer.Option(
        ...,
        "--output-dir",
        "-o",
        readable=True,
        writable=True,
        dir_okay=True,
        file_okay=False,
    ),
):
    if channels is not None:
        assert len(channels) > 0, "At least one channel is required"

    if reference_range is None:
        reference_range = (7000, 8000)

    if smooth_window is None:
        smooth_window = 250

    rayleigh_fit_from_file(
        file=file,
        initial_hour=initial_hour,
        duration=duration,
        channels=channels,
        reference_range=reference_range,
        smooth_window=smooth_window,
        output_dir=output_dir,
        save_fig=save_fig,
    )



import pytz
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
import smtplib
from email import utils

from loguru import logger

from gfatpy.lidar.utils.utils import LIDAR_INFO


def send_email(email_sender: dict, email_receiver: list[str], subject: str, email_content: str):
    """Send email to the specified email addresses

    Args:
        email_sender (dict): Dictionary with following keys: server, port, email_name, password.
        email_receiver (list[str]): List of email addresses to send the email to.
        email_content (str): Content of the email to be sent.

    Returns:
        _type_: _description_
    """
    
    # Check Input
    if isinstance(email_receiver, str):
        email_receiver = [email_receiver]

    logger.info("Send Email")    
    
    return_code = 1
    try:
        # Prepare Message
        message = MIMEMultipart()
        message["To"] = ", ".join(email_receiver)
        message["To"] = ", ".join(email_receiver)
        message["Date"] = formatdate(localtime=True)
        message["Subject"] = subject
        message.attach(MIMEText(email_content))

        # Connect to email server and send email
        # ssl_context = ssl.create_default_context()
        # conn = smtplib.SMTP_SSL(email_sender['server'], email_sender['port'], context=ssl_context)
        conn = smtplib.SMTP(email_sender["server"], email_sender["port"])
        conn.ehlo()
        conn.starttls()
        conn.ehlo()
        conn.login(email_sender["email_name"], email_sender["password"])
        conn.sendmail(email_sender["email_name"], email_receiver, message.as_string())
        logger.info("Email Sent")
    except:
        logger.error("Email not sent")
        return_code = 0
    finally:
        conn.quit()
    return return_code


# from curses import set_tabsize
from pathlib import Path
from loguru import logger
import numpy as np
import xarray as xr


# from gfatpy.lidar.preprocessing.lidar_gluing_bravo_aranda import gluing
from gfatpy.utils import calibration
from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.utils.io import read_yaml_from_info
from gfatpy.lidar.preprocessing.gluing_de_la_rosa_slope import gluing
from gfatpy.utils.utils import adaptive_moving_average, sliding_average
from gfatpy.lidar.utils.file_manager import (
    add_required_channels,
    filename2info,
    search_dc,
)
from gfatpy.lidar.preprocessing.lidar_preprocessing_tools import (
    ff_2D_overlap_from_channels,
)

# from .lidar_merge import apply_polarization_merge
# from .lidar_preprocessing_tools import *  # TODO: Remove wildcard importations within modules

# warnings.filterwarnings("ignore")

__author__ = "Bravo-Aranda, Juan Antonio"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Bravo-Aranda, Juan Antonio"
__email__ = "jabravo@ugr.es"
__status__ = "Production"


""" DEFAULT AUXILIAR INFO
"""
# Root Directory (in NASGFAT)  according to operative system
""" LIDAR PREPROCESSING
"""


def preprocess(
    file_or_path: Path | str,
    channels: list[str] | None = None,
    crop_ranges: tuple[float, float] | None = (0, 20000),
    background_ranges: tuple[float, float] | None = None,
    apply_sm: bool = False,
    apply_dc: bool = True,
    apply_dt: bool = True,
    apply_bg: bool = True,
    apply_bz: bool = True,
    apply_ov: bool = False,
    save_dc: bool = False,
    save_bg: bool = False,
    gluing_products: bool = False,
    force_dc_in_session=False,
    overlap_path: Path | str | None = None,
    smooth_mode: str | None = None,
    **kwargs,
) -> xr.Dataset:
    """It preprocesses lidar netcdf file. It can optionally apply dark-measurement, deadtime, background, zero-bin, and overlap correction and calculate gluing products.

    Args:
        # ...existing args...
        -apply_sliding_average (bool, optional): it applies a sliding average to the signal variables. Defaults to False.
        -y_sm_lims (tuple[float, float], optional): limits for the sliding average. Defaults to (0, 20000).
        -y_sm_hwin (tuple[float, float], optional): half window size for the sliding average. Defaults to (100, 1000).


    Returns:
        -xr.Dataset: xarray.Dataset with the lidar data and information.
    """

    _p = Path(file_or_path)
    if not _p.exists():
        raise ValueError(f"{file_or_path} not found")

    # Get info from filename
    lidar_nick, _, _, _, _, date = filename2info(_p.name)
    global INFO
    INFO = read_yaml_from_info(lidar_nick, date)

    with xr.open_dataset(_p, chunks={}, engine="netcdf4") as _nc:
        dataset = _nc

    if channels is None:
        channels = dataset.channel.values.tolist()
        raw_channels = channels
        product_channels = []
    else:
        channels = add_required_channels(lidar_nick, channels, date)
        # Split channels by raw/products
        raw_channels = [
            channel_
            for channel_ in channels
            if channel_ not in INFO["product_channels"].keys()
        ]

        dataset = drop_unwanted_channels(dataset, keep_channels=raw_channels)

    # If header is defined in INFO, its fields are used to update the dataset. Variables and attributes can be updated. See example on info_alh_20221117.yml
    if "header" in INFO.keys():
        update_from_info(dataset, INFO["header"])

    if apply_sm:
        dataset = apply_smooth(dataset, smooth_mode=smooth_mode, **kwargs)

    if apply_dc:
        dataset = apply_dark_current_correction(
            dataset,
            rs_path=_p,
            save_dc=save_dc,
            force_dc_in_session=force_dc_in_session,
            apply_sm=apply_sm,
            smooth_mode=smooth_mode,
            **kwargs
        )
    else:
        dataset.attrs["dc_corrected"] = str(False)

    if apply_dt:
        dataset = apply_dead_time_correction(
            dataset
        )  # TODO implement search_dt() in the same way of search_dc
    else:
        dataset.attrs["dt_corrected"] = str(False)

    if apply_bg:
        dataset = apply_background_correction(
            dataset, background_ranges=background_ranges, save_bg=save_bg
        )
    else:
        dataset.attrs["bg_corrected"] = str(False)

    if apply_bz:
        dataset = apply_bin_zero_correction(dataset, rs_path=_p)
    else:
        dataset.attrs["bz_corrected"] = str(False)

    if apply_ov:
        dataset = apply_overlap_correction(dataset, overlap_path)
        dataset.attrs["ov_corrected"] = "True"
    else:
        dataset.attrs["ov_corrected"] = "False"

    dataset = apply_crop_ranges_correction(dataset, crop_ranges=crop_ranges)

    if gluing_products:
        dataset = apply_detection_mode_merge(dataset)

    dataset = add_height(dataset)

    return dataset


def update_from_info(dataset: xr.Dataset, header: dict):
    """Update dataset variables and attributes from INFO header.

    Args:

        - dataset (xr.Dataset): Dataset to be updated.
        - header (dict): Header from INFO.

    Raises:

        - RuntimeError: Variables or attributes could not be updated.
        - RuntimeError: Attributes could not be updated.

    Returns:

        - xr.Dataset: Updated dataset.
    """
    for key_ in header.keys():
        if key_ in [*dataset.variables.keys()]:
            try:
                # Replace dataset[key_] with header[key_] using the same dimensions and coordinates
                dataset[key_] = xr.DataArray(
                    header[key_], dims=dataset[key_].dims, coords=dataset[key_].coords
                )
            except Exception as e:
                raise RuntimeError(
                    f"Could not update variable {key_} from INFO. Error: {e}"
                )
        if key_ in [*dataset.attrs.keys()]:
            try:
                dataset.attrs[key_] = header[key_]
            except Exception as e:
                raise RuntimeError(
                    f"Could not update attribute {key_} from INFO. Error: {e}"
                )
    return dataset


def drop_unwanted_channels(
    dataset: xr.Dataset, keep_channels: list[str] | None
) -> xr.Dataset:
    """Drop unwanted channels from dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be updated.
        - keep_channels (list[str] | None): List of channels to be removed. Defaults to None meaning all channels are kept.

    Returns:

        - xr.Dataset: Updated dataset.
    """
    if keep_channels is None:
        return dataset
    remove_channels = list(
        filter(lambda channel_: channel_ not in keep_channels, dataset.channel.values)
    )
    remove_signals = list(
        f"signal_{channel_}"
        for channel_ in remove_channels
        if f"signal_{channel_}" in dataset.keys()
    )
    dataset = dataset.drop_vars(remove_signals, errors="raise")
    remove_stderr = list(
        f"stderr_{channel_}"
        for channel_ in remove_channels
        if f"stderr_{channel_}" in dataset.keys()
    )

    dataset = dataset.drop_vars(remove_stderr, errors="raise")
    dataset = dataset.sel(channel=keep_channels)

    return dataset


def add_height(dataset: xr.Dataset) -> xr.Dataset:
    """Add height coordinate to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be updated.

    Raises:

        - RuntimeError: Zenithal angle is not constant in time. Cannot add height coordinate.

    Returns:

        - xr.Dataset: Updated dataset.
    """
    # Add height because of zenithal angle
    if "zenithal_angle" in [*dataset.variables.keys()]:
        if dataset["zenithal_angle"].values.all():
            zenithal_angle = np.deg2rad(dataset["zenithal_angle"].values[0])
            if zenithal_angle != 0:
                dataset["height"] = dataset["range"] * np.cos(zenithal_angle)
            else:
                dataset["height"] = dataset["range"].copy()
        else:
            raise RuntimeError(
                "Zenithal angle is not constant in time. Cannot add height coordinate."
            )
    return dataset


def apply_smooth(
    dataset: xr.Dataset, smooth_mode: str | None = None, **kwargs
) -> xr.Dataset:
    # Apply moving average if requested
    if smooth_mode == "moving":
        logger.info(f"Applying moving average smoothing...")
        for channel_ in dataset.channel.values:  # type: ignore
            dataset[f"signal_{channel_}"] = xr.DataArray(
                adaptive_moving_average(
                    dataset[f"signal_{channel_}"].values,
                    window_sizes=kwargs.get("window_sizes", 10.0),
                ),
                dims=dataset[f"signal_{channel_}"].dims,
                coords=dataset[f"signal_{channel_}"].coords,
            )
    elif smooth_mode == "sliding":
        logger.info(f"Applying sliding average smoothing...")
        for channel_ in dataset.channel.values:
            dataset[f"signal_{channel_}"] = sliding_average(
                dataset[f"signal_{channel_}"],
                maximum_range=kwargs.get("sliding_maximum_range", 4000.0),
                window_range=kwargs.get("window_range", (10, 300)),
            )
    elif smooth_mode == "binning":
        logger.info(f"Not implemented yet...")
        # logger.info(f"Applying binning average smoothing...")
        # bin_size = kwargs.get("binning_average_bin", 10)
        # binned_range = np.arange(dataset["range"].min(), dataset["range"].max() + bin_size, bin_size)
        # # # Perform binning and mean computation lazily
        # dataset = dataset.groupby_bins(
        #     "range", binned_range, eagerly_compute_group=False
        # ).mean()
        # dataset = dataset.rename({f"range_bins": "range"}).assign_coords({"range": binned_range[:-1]})
    elif smooth_mode is None:
        logger.info(f"No smoothing applied...")
    else:
        raise ValueError(
            "smooth_mode not recognized [choose: moving, sliding, binning]."
        )
    return dataset


def apply_dark_current_correction(
    dataset: xr.Dataset,
    rs_path: Path,
    save_dc: bool = False,
    force_dc_in_session=False,
    apply_sm: bool = False,
    smooth_mode: str | None = None,
    crop_ranges: tuple[float, float] | None = (0, 20000),
    **kwargs,
) -> xr.Dataset:
    """Apply dark current correction to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be corrected.
        - rs_path (Path): Path to the raw data.
        - save_dc (bool, optional): Flag to save the dark current in the dataset. Defaults to False.
        - force_dc_in_session (bool, optional): Flag to force the use of the dark current linked to this measurement period. If not found, the correction cannot be performed. Defaults to False.

    Raises:

        - Warning: Dark current std is too high for an specified channel. Dark measurement shall be not appropiated.

    Returns:

        - xr.Dataset: Dataset with dark current corrected.
    """
    groups = calibration.split_continous_measurements(dataset.time.values)
    channels = dataset.channel.values

    analog_channels: list[str] = list(filter(lambda c: c.endswith("a"), channels))

    for group in groups:
        dc_path = search_dc(
            rs_path,
            session_period=group[[0, -1]],
            force_dc_in_session=force_dc_in_session,
        )
        dc = xr.open_dataset(dc_path)

        dc = drop_unwanted_channels(dc, keep_channels=analog_channels)

        if apply_sm:
            dc = apply_smooth(dc, smooth_mode=smooth_mode, **kwargs)
            
        lower_idx = np.where(dataset.time == group[0])[0][0]
        upper_idx = np.where(dataset.time == group[-1])[0][0] + 1

        for channel in analog_channels:
            signal_str = f"signal_{channel}"
            if dc[signal_str].sel(range=slice(0, 5000)).std("range").std("time") > 0.07:
                raise Warning(
                    f"Dark current std is too high for channel {channel} in {dc_path}. Check if DC was measured correctly."
                )

            dc_mean = dc[signal_str].mean(axis=0)
            
            dataset[signal_str].loc[dict(time=group)] -= dc_mean.values[np.newaxis, :]
            # dataset[signal_str].values[:,lower_idx:upper_idx] -= dc_mean.values #TODO: too slow

            if save_dc:
                if (
                    f"dc_{channel}" not in list(dataset.variables.keys())
                    and lower_idx == 0
                ):
                    dataset[f"dc_{channel}"] = dataset[signal_str] * np.nan

                dataset[f"dc_{channel}"][lower_idx:upper_idx] = dc_mean

    dataset.attrs["dc_corrected"] = str(True)

    return dataset


def apply_dead_time_correction(dataset: xr.Dataset) -> xr.Dataset:
    """Apply dead time correction to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be corrected.

    Raises:

        - ValueError: No dead time value defined in INFO->{lidar_name}->{channel}.
        - ValueError: No dead time value defined in INFO->{lidar_name}.

    Returns:

        - xr.Dataset: Dataset with dead time corrected.
    """
    # dt_path = search_dt(rs_path, session_period=dataset.time.values[[0,-1]]) #
    # dt_dict = open_dataset(dt_path) #TODO

    lidar_name = dataset.attrs["system"].lower()
    try:
        dt_dict = {
            key: value["dead_time_ns"]
            for (key, value) in INFO["channels"].items()
            if value.get("dead_time_ns", False)
        }
    except Exception:
        raise ValueError(f"No dead time value defined in INFO->{lidar_name}].")

    photocounting_channels: list[str] = list(
        filter(lambda c: c.endswith("p"), dataset.channel.values)
    )

    for channel in photocounting_channels:
        # tau from ns to us
        try:
            tau_us = dt_dict[channel] * 1e-3
        except KeyError:
            raise ValueError(  # TODO
                f"No dead time value defined in INFO->{lidar_name}->{channel}."
            )

        signal_str = f"signal_{channel}"

        # Eq 4 [D'Amico et al., 2016]
        # No infinites nor negative values
        # condition = np.logical_and(~np.isinf(dataset[signal_str]), dataset[signal_str] > 0)
        # dataset[signal_str] = dataset[signal_str].where(condition)/ ( 1 - dataset[signal_str].where(condition) * tau_us )

        dataset[signal_str] = dataset[signal_str] / (1 - dataset[signal_str] * tau_us)
        dataset.attrs["dt_corrected"] = str(True)

    return dataset


def apply_background_correction(
    dataset: xr.Dataset,
    background_ranges: tuple[float, float] | None = None,
    save_bg: bool = False,
) -> xr.Dataset:
    """Apply background correction to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be corrected.
        - background_ranges (tuple[float, float] | None, optional): Range where background average will be performed for each signal. This value is subtracted to the whole profile. Defaults to None.
        - save_bg (bool, optional): Flag to save the background in the dataset. Defaults to False.

    Raises:

        - ValueError: background_ranges should be in order (min, max).

    Returns:

        - xr.Dataset: Dataset with background corrected.
    """
    if background_ranges is None:
        background_ranges = (
            dataset.attrs["BCK_MIN_ALT"],
            dataset.attrs["BCK_MAX_ALT"],
        )

    if background_ranges[1] <= background_ranges[0]:
        raise ValueError("background_ranges should be in order (min, max)")

    ranges_between = (background_ranges[0] < dataset.range) & (
        dataset.range < background_ranges[1]
    )
    channels: list[str] = dataset.channel.values
    for channel in channels:
        signal_str = f"signal_{channel}"
        try:
            background = dataset[signal_str].loc[:, ranges_between].mean(axis=1)
        except:
            background = np.ones(1)
        dataset[signal_str] -= background

        if save_bg:
            dataset[f"bg_{channel}"] = background

    dataset.attrs["bg_corrected"] = str(True)

    return dataset


def apply_bin_zero_correction(dataset: xr.Dataset, rs_path: Path) -> xr.Dataset:
    """Apply zero bin correction to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be corrected.
        - rs_path (Path): Path to the raw data.

    Raises:

        - ValueError: No bin zero value defined in INFO->{lidar_name}.

    Returns:

        - xr.Dataset: Dataset with zero bin corrected.
    """
    # bz_path = search_bz(rs_path, session_period=dataset.time.values[[0,-1]]) #
    # bz_dict = open_dataset(bz_path) #TODO
    bz_dict = None

    if bz_dict is None:
        lidar_name = dataset.attrs["system"].lower()
        try:
            bz_dict = {
                key: value["bin_zero"] for (key, value) in INFO["channels"].items()
            }
        except Exception:
            raise ValueError(f"No bin zero value defined in INFO->{lidar_name}.")

    channels: list[str] = dataset.channel.values
    for channel in channels:
        signal_str = f"signal_{channel}"
        dataset[signal_str] = dataset[signal_str].shift(
            range=-bz_dict[channel], fill_value=0.0
        )

    dataset.attrs["bz_corrected"] = str(True)

    return dataset


def apply_overlap_correction(
    dataset: xr.Dataset,
    ff_overlap_path: Path | str | None = None,
    nf_overlap_path: Path | str | None = None,
) -> xr.Dataset:
    """Apply overlap correction to dataset.

    Args:

        - dataset (xr.Dataset): Dataset to be corrected.
        - overlap_path (Path): Path to the overlap file.

    Returns:

        - xr.Dataset: Dataset with overlap corrected.
    """

    def apply_correction(
        dataset: xr.Dataset,
        overlap: xr.DataArray,
        channels2correct: list[str],
        overlap_channel: str,
    ) -> xr.Dataset:

        try:
            for channel_ in channels2correct:
                if channel_ in dataset.channel.values:
                    signal_str = f"signal_{channel_}"

                    # Perform overlap correction
                    dataset[signal_str] = dataset[signal_str] / overlap
                    dataset[signal_str].attrs["overlap_applied"] = overlap_channel
                    dataset["overlap_corrected"][dataset.channel.values == channel_] = 1
        except:
            raise Warning("Could not apply overlap correction")
        return dataset

    lidar_name = dataset.attrs["system"].lower()
    overlap_channels = INFO["overlap_channels"]
    linked_channels = INFO["overlap_linked_channels"]

    # New array to store if overlap correction was applied
    overlap_corrected = xr.DataArray(
        np.zeros(dataset.channel.size),
        dims=("channel"),
        coords={"channel": dataset.channel},
    )
    dataset["overlap_corrected"] = overlap_corrected

    # Apply overlap correction for far-field channels
    # TODO: Move to LIDAR_INFO

    # Load overlap dataset
    if ff_overlap_path is not None:
        overlap = xr.open_dataarray(ff_overlap_path)
    else:
        overlap = None

    for channel_ in linked_channels.keys():
        # Check channel_ in dataset

        if channel_ not in dataset.channel.values:
            continue

        # Select overlap array for channel
        if overlap is not None:
            if channel_ in overlap.channel.values:
                overlap_ = overlap.sel(channel=channel_)
            else:
                raise ValueError(f"Channel {channel_} not found in overlap file")
        else:
            nf_channel_ = overlap_channels[channel_]
            if not nf_channel_ in dataset.channel.values:
                raise ValueError(
                    f"Channel {nf_channel_} not found in dataset. Please include near-field channel in dataset to retrieve overlap."
                )
            overlap_ = ff_2D_overlap_from_channels(
                dataset, channel_ff=channel_, channel_nf=nf_channel_
            )
        # Check size of arrays
        if dataset.range.size != overlap_.range.size:
            # Adapt size of overlap array to dataset
            overlap_ = overlap_.interp(range=dataset.range, method="linear")

        # Save overlap array
        if overlap_.values.any():
            # Apply overlap correction
            dataset = apply_correction(
                dataset, overlap_, linked_channels[channel_], overlap_channel=channel_
            )

            dataset[f"overlap_{channel_}"] = overlap_

    # Apply overlap correction for far-field channels
    if nf_overlap_path is not None:
        linked_channels = {
            "355npa": [
                "355npa",
                "355npp",
                "355nsa",
                "355nsp",
                "387nta",
                "387ntp",
                "408nta",
                "408ntp",
            ],
            "532npa": ["532npa", "532npp", "607nta", "607ntp"],
        }
        # Load overlap dataset
        overlap = xr.open_dataarray(nf_overlap_path)

        for channel_ in overlap.channel.values:
            # Select overlap array for channel
            overlap_ = overlap.sel(channel=channel_)

            # Check size of arrays
            if dataset.range.size != overlap_.range.size:
                # Adapt size of overlap array to dataset
                overlap_ = overlap_.interp(range=dataset.range, method="linear")

            # Apply overlap correction
            dataset = apply_correction(
                dataset, overlap_, linked_channels[channel_], overlap_channel=channel_
            )

            # Save overlap array
            if dataset["overlap_corrected"].values.any():
                dataset[f"overlap_{channel_}"] = overlap_
    return dataset


def apply_crop_ranges_correction(
    dataset: xr.Dataset, crop_ranges: tuple[float, float] | None = (0, 20000)
) -> xr.Dataset:
    """It crops the dataset range dimension to `crop_ranges` values.

    Args:

        - dataset (xr.Dataset): Dataset to be cropped.
        - crop_ranges (tuple[float, float] | None, optional): Mininum and maximum ranges. Defaults to (0, 20000).

    Raises:

        - ValueError: crop_ranges should be in order (min, max).

    Returns:

        - xr.Dataset: Dataset with range cropped.
    """
    # TODO: Apply crop ranges. With dataset.sel(range=slice(*crop_ranges))

    if crop_ranges is None:
        return dataset

    if crop_ranges[0] > crop_ranges[-1]:
        raise ValueError("crop_ranges should be in order (min, max)")

    dataset = dataset.sel(range=slice(*crop_ranges))

    return dataset


def apply_detection_mode_merge(dataset: xr.Dataset) -> xr.Dataset:
    """It performs the detection-mode gluing.

    Args:
        dataset (xr.Dataset): Dataset to be merged.

    Returns:
        xr.Dataset: Merged dataset.
    """
    LIDAR_INFO["metadata"]["code_telescope_str2number"]
    LIDAR_INFO["metadata"]["code_mode_str2number"]

    channels_pc: list[str] = list(
        filter(
            lambda c: any(
                c.startswith(vc) and c.endswith("p") for vc in dataset.channel.values
            ),
            dataset.channel.values,
        )
    )

    range_m = dataset["range"].values
    glued_list: list[dict] = []

    for channel_pc in channels_pc:
        channel_an = f"{channel_pc[0:-1]}a"

        if f"signal_{channel_an}" not in list(dataset.variables.keys()):
            continue

        signal_gl = gluing(
            dataset[f"signal_{channel_an}"], dataset[f"signal_{channel_pc}"]
        )

        print(f"Gluing done for channels: {channel_an} and {channel_pc}")

        polarization = channel_pc[-1] if channel_pc[-1] in ["p", "s"] else ""
        telescope = channel_pc[0]  # Primera letra representa el telescopio
        wavelength = channel_pc[1:-1]  # Longitud de onda sin la primera ni última letra

        glued_list.append(
            {
                "name": f"{channel_pc[0:-1]}g",
                "signal": signal_gl,
                "polarization": polarization,
                "telescope": telescope,
                "wavelength": wavelength,
            }
        )

    glued_signal = {
        f"signal_{glued['name']}": (["time", "range"], glued["signal"])
        for glued in glued_list
    }

    other_var = {
        "wavelength": (["channel"], [g["wavelength"] for g in glued_list]),
        "polarization": (["channel"], [g["polarization"] for g in glued_list]),
        "telescope": (["channel"], [g["telescope"] for g in glued_list]),
        "bin_shift": (["channel"], [0 for _ in glued_list]),
    }

    glued_dataset = xr.Dataset(
        glued_signal | other_var,
        coords={
            "range": range_m,
            "time": dataset["time"].values,
            "channel": list(map(lambda i: i["name"], glued_list)),
        },
    )

    return xr.merge([dataset, glued_dataset])




from pathlib import Path
import numpy as np
import pandas as pd
import xarray as xr
from datetime import datetime

from gfatpy.lidar.nc_convert.measurement import Measurement
from gfatpy.generalife.licel import LicelReader

DZ = 7.5
MAX_BINS = 2000

def licel_to_datetime(licel_name: str) -> datetime | None:
    """
    Convert a Licel file name to a datetime object.

    The Licel file name is expected to have a specific format where the last 7 characters
    before the file extension represent the date and time in the following way:
    - The last 2 characters represent the hour and minute.
    - The 3rd and 4th characters from the end represent the day.
    - The 5th character from the end is a hexadecimal digit representing the month.
    - The 6th and 7th characters from the end represent the year.

    Args:
        licel_name (str): The name of the Licel file, including its extension.

    Returns:
        datetime | None: A datetime object representing the date and time extracted from the file name,
                         or None if the file name format is unexpected.
    """
    name, extension = licel_name.split(".")
    try:
        month_decimal = int(name[-5], 16)
        return datetime.strptime(f"{name[-7:-5]}{month_decimal:02d}{name[-4:-2]}T{name[-2:]}{extension[:4]}", r"%y%m%dT%H%M%S")
    except ValueError:
        print(f"Skipping file {licel_name} due to unexpected name format.")
        return None
    

def measurement_to_nc(measurement: Measurement, output_dir: Path):    
    """Converts measurement data to a NetCDF file. The function reads data from the measurement object, processes it, and saves it as a NetCDF file in the specified output directory. The NetCDF file contains raw data, voltages, and metadata attributes such as system name, location, latitude, longitude, altitude, frequency, bin width, number of bins, number of channels, center wavelength, discriminator, bin_corrected, and bcg_corrected.

    Args:
        measurement (Measurement): The measurement object containing data to be converted.
        output_dir (Path): The directory where the output NetCDF file will be saved.

    Raises:
        ValueError: If no files are found in the measurement object.

    """
    files = measurement.get_filepaths()
    
    if files is None or len(files) == 0:
        raise ValueError("No files found.")
    
    times = [licel_to_datetime(file.name) for file in files]
    
    df = pd.read_csv([*files][0].absolute().as_posix(), header=None, skiprows=3, delimiter=' ', encoding='ISO-8859-1', on_bad_lines='skip')
    wavelengths = df[8].dropna().astype(float).values
    voltages = df[6].dropna().astype(float).values

    data = np.empty((len(files), len(wavelengths), MAX_BINS), dtype=float)

    for file_idx, file_path in enumerate(files):
        sp = LicelReader.LicelFileReader(file_path)
        for i, _ in enumerate(wavelengths):
            data[file_idx, i, :] = sp.dataSet[i].physData

    # Create an xarray DataArray from the data array
    raw = xr.DataArray(
        data,
        coords={'wavelength': wavelengths, 'times': times, 'range': DZ*np.arange(MAX_BINS)},
        dims=['times', 'wavelength', 'range']
    )

    voltages = xr.DataArray(
        voltages,
        coords={'wavelength': wavelengths},
        dims=['wavelength']
    )

    center_wavelength = (wavelengths[16] + wavelengths[15]) / 2

    # Create an xarray Dataset from the DataArray
    ds = xr.Dataset({'raw': raw, 'voltages': voltages})

    ds.attrs.update({
            'system name': 'Generalife',  
            'location': 'Granada',
            'latitude': '37.1641° N',
            'longitude': '3.6026° W',
            'altitude': '680 m',
            'frequency': '10 Hz',
            'bin width': '7.5 m',
            'number of bins': '2000',
            'number of channels': '32',
            'center wavelength': center_wavelength,            
            'discriminator' : '4.0',
            'bin_corrected': 'False',
            'bcg_corrected': 'False',
        })

    if isinstance(times[0], datetime):
        datestr = times[0].strftime('%Y%m%d')   
        filename = f"gnl_1a_Prs_rs_ff_{datestr}.nc"
        output_path = output_dir / "generalife" / "1a" / f"{times[0].strftime('%Y')}" / f"{times[0].strftime('%m')}" / f"{times[0].strftime('%d')}" / filename
        ds.to_netcdf(output_path)




import numpy as np
from pathlib import Path
from datetime import datetime
from matplotlib.lines import Line2D
from matplotlib.figure import Figure
import xarray as xr
from gfatpy.radar.rpg_nc import rpg
from gfatpy.radar.utils import check_is_netcdf
from gfatpy.utils.utils import parse_datetime


class nebula():
    def __init__(self, ka_nc_path: Path, ww_nc_path: Path):
        """
        Initializes the Nebula class with the given file paths for ka and ww netCDF files.

        Args:
            ka_nc_path (Path): Path to the ka netCDF file.
            ww_nc_path (Path): Path to the ww netCDF file.
        """
        self.paths = [check_is_netcdf(ka_nc_path), check_is_netcdf(ww_nc_path)]
        self.ka = rpg(ka_nc_path)
        self.ww = rpg(ww_nc_path)
        self._data = None
        self._band = None
    
    @property
    def level(self) -> int:
        if self.ka.level == self.ww.level:
            return self.ka.level
        else:
            raise ValueError("Different type radar measurement cannot be combined.")

    @property
    def type(self) -> str:
        if self.ka.type == self.ww.type:
            return self.ka.type
        else:
            raise ValueError("Different type radar measurement cannot be combined.")

    @property
    def data(self) -> xr.Dataset:
        if self._data is None:
            if self.level == 1:         
                _data = self.ka.data.copy()                   
                _data = _data.drop_vars(_data.variables.keys())

                _data["DWR"] = 10 * np.log10(self.ka.data["Ze"] / self.ww.data["Ze"])
                _data["DWR"].attrs["units"] = "dB"
                _data["DWR"].attrs["long_name"] = "Ka-W DWR"

                _data["DDV"] = self.ka.data["v"] - self.ww.data["v"]
                _data["DDV"].attrs["units"] = "m/s"
                _data["DDV"].attrs["long_name"] = "Ka-W DDV"
            elif self.level == 0:
                pass
                # breakpoint()
                # _data = self.ka.data.copy()
                # _data["DWR"] = 10 * np.log10(self.ka.data["Ze"] / self.ww.data["Ze"])
                # _data["DWR"].attrs["units"] = "dB"
                # _data["DWR"].attrs["long_name"] = "Ka-W DWR"

                # _data["DDV"] = self.ka.data["v"] - self.ww.data["v"]
                # _data["DDV"].attrs["units"] = "m/s"
                # _data["DDV"].attrs["long_name"] = "Ka-W DDV"
            return _data
        else:
            return self._data        

    def quicklook(self, variable: str | list[str], **kwargs) -> tuple[list[Figure] | None, list[Path] | None]:
        """
        Generate quicklook plots for the specified variables.
        Parameters:
        -----------
        variable : str or list of str
            The variable(s) to generate quicklook plots for. Must be a key or keys in self.data.
        **kwargs : dict
            Additional keyword arguments to pass to the plotting function.
        Returns:
        --------
        tuple[list[Figure] | None, list[Path] | None]
            A tuple containing a list of generated figures and a list of paths where the figures are saved.
            If no figures are generated, both lists will be None.
        """        
        nb_variables = [var for var in variable if var in self.data.keys()]

        #Create a fake NEBULA rpg instatnce to plot
        nb_rpg_class = rpg(self.ka.path)
        nb_rpg_class._data = self.data #replace ka data with nb data 
        nb_rpg_class._band = 'Dual'

        figures, paths = nb_rpg_class.quicklook(nb_variables, **kwargs)

        return figures, paths

    def plot_profile(
        self,
        variable: str | list[str],
        target_time: datetime | np.datetime64,
        range_limits: tuple[float, float],
        **kwargs,
    ) -> tuple[list[Figure] | None, list[Path] | None]:
        """
        Plots the profile of the specified variable(s) over the given time and range limits.
        Parameters:
        -----------
        target_time : datetime | np.datetime64
            The target time for which the profile is to be plotted.
        range_limits : tuple[float, float]
        variable : str | list[str]
            The variable(s) to be plotted. If None, all variables will be plotted.
            The range limits for the profile plot.
        **kwargs : dict
            Additional keyword arguments to customize the plot. Possible keys include:
            - savefig (bool): Whether to save the figure. Default is False.
            - ka_color (str): Color for the Ka-band plot. Default is "red".
            - w_color (str): Color for the W-band plot. Default is "blue".
            - output_dir (Path): Directory to save the figure if savefig is True. Default is current working directory.
            - dpi (int): Dots per inch for the saved figure. Default is 300.
        Returns:
        --------
        tuple[Figure, Path | None]
            The figure object and the path to the saved figure (if savefig is True), otherwise None.
        """
        
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]
        
        list_of_figs, list_of_paths = [], []

        if isinstance(variable, str):
            variables = [variable]
        elif isinstance(variable, list):
            variables = variable
        
        ka_ww_variables = [var for var in variables if (var in self.ka.data.keys() and var in self.ww.data.keys())]

        for variable_ in ka_ww_variables:
            fig1, _ = self.ka.plot_profile(
                target_times=target_time,
                range_limits=range_limits,
                variable=variable_,
                **_kwargs,
            )

            _kwargs["fig"] = fig1
            _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
            
            fig, _ = self.ww.plot_profile(
                target_times=target_time,
                range_limits=range_limits,
                variable=variable_,
                **_kwargs,
            )

            ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[1]
            ka_line.set_label(f"Ka-band")
            w_line.set_label(f"W-band")

            ax = fig.get_axes()[0]
            ax.legend()

            if kwargs.get("savefig", False):
                filepath = (
                    kwargs.get("output_dir", Path.cwd())
                    / f"profile_{variable_}_{target_time:%Y%m%dT%H%M%S}.png"
                )
                fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
                list_of_figs.append(fig)
                list_of_paths.append(filepath) 
       
        #Plot dual products 
        #Create a fake NEBULA rpg instatnce to plot
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        nb_rpg_class = rpg(self.ka.path)
        nb_rpg_class._data = self.data #replace ka data with nb data 
        nb_rpg_class._band = 'Dual'
        
        nb_variables = [var for var in variables if var in self.data.keys()]        
        for variable_ in nb_variables:
            fig, _ = nb_rpg_class.plot_profile(
                target_times=target_time,
                range_limits=range_limits,
                variable=variable_,
                **_kwargs,
            )
            
            if kwargs.get("savefig", False):
                filepath = (
                    kwargs.get("output_dir", Path.cwd())
                    / f"profile_{variable_}_{target_time:%Y%m%dT%H%M%S}.png"
                )
                fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
                list_of_figs.append(fig)
                list_of_paths.append(filepath) 
        if len(list_of_figs)==0 and len(list_of_paths) == 0:
            list_of_figs, list_of_paths = None, None
        
        return list_of_figs, list_of_paths

    def plot_timeseries(
        self,
        variable: str | list[str],
        target_range: float,
        time_limits: tuple[datetime | np.datetime64, datetime | np.datetime64] | None = None,
        **kwargs,
    ) -> tuple[list[Figure] | None, list[Path] | None]:
        """
        Plots a time series for the specified target range and variable.
        Parameters:
        -----------
        target_range : float
            The target range for which the time series is to be plotted.
        time_limits : tuple[datetime | np.datetime64, datetime | np.datetime64] | None, optional
            A tuple specifying the start and end times for the time series plot. If None, the full time range of the data is used.
        variable : str | None, optional
            The variable to be plotted. If None, a default variable is used.
        **kwargs : dict
            Additional keyword arguments to customize the plot. Common options include:
            - savefig (bool): Whether to save the figure to a file. Default is False.
            - ka_color (str): Color for the Ka-band plot. Default is "red".
            - w_color (str): Color for the W-band plot. Default is "blue".
            - output_dir (Path): Directory to save the figure if savefig is True. Default is the current working directory.
            - dpi (int): Dots per inch for the saved figure. Default is 300.
        Returns:
        --------
        tuple[Figure, Path | None]
            A tuple containing the figure object and the file path where the figure is saved (if applicable).
        """
        
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False #Not directy save because modifications are required.
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]
        list_of_figs, list_of_paths = [], []

        if isinstance(variable, str):
            variables = [variable]
        elif isinstance(variable, list):
            variables = variable

        if  time_limits is None:
            time_limits = (parse_datetime(self.ka.data.time[0].values), parse_datetime(self.ka.data.time[-1].values))
        
        ka_ww_variables = [var for var in variables if (var in self.ka.data.keys() and var in self.ww.data.keys())]
                
        for variable in ka_ww_variables:
            fig1, _ = self.ka.plot_timeseries(
                target_ranges=target_range,
                time_limits=time_limits,
                variable=variable,
                **_kwargs,
            )

            _kwargs["fig"] = fig1
            _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
            
            fig, _ = self.ww.plot_timeseries(
                target_ranges=target_range,
                time_limits=time_limits,
                variable=variable,
                **_kwargs,
            )
            
            ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[1]
            ka_line.set_label(f"Ka-band")
            w_line.set_label(f"W-band")

            ax = fig.get_axes()[0]
            variable_string = self.ka.data[variable].attrs['long_name'].replace('W-band','').replace("Ka-band",'')
            ax.set_ylabel(f"{variable_string}, [{self.ka.data[variable].attrs['units']}]")
            ax.legend()

            if kwargs.get("savefig", False):
                filepath = (
                    kwargs.get("output_dir", Path.cwd())
                    / f"timeseries_{variable}_{time_limits[0]:%Y%m%dT%H%M%S}_{time_limits[-1]:%Y%m%dT%H%M%S}_{target_range}.png"
                )
                fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
                list_of_figs.append(fig)
                list_of_paths.append(filepath)

        #Plot dual products 
        #Create a fake NEBULA rpg instatnce to plot
        nb_rpg_class = rpg(self.ka.path)
        nb_rpg_class._data = self.data #replace ka data with nb data 
        nb_rpg_class._band = 'Dual'
        
        nb_variables = [var for var in variables if var in self.data.keys()]   
        _kwargs = kwargs.copy()
        _kwargs['savefig'] = False
        for variable_ in nb_variables:            
            fig_, _ = nb_rpg_class.plot_timeseries(
                target_ranges=target_range,
                time_limits=time_limits,
                variable=variable_,
                **_kwargs, #not directly used because no needs to savefig here.
            )

            if kwargs.get("savefig", False):
                filepath_ = (
                    kwargs.get("output_dir", Path.cwd())
                    / f"timeseries_{variable_}_{time_limits[0]:%Y%m%dT%H%M%S}_{time_limits[-1]:%Y%m%dT%H%M%S}_{target_range}.png"
                )                
                fig_.savefig(filepath_, dpi=kwargs.get("dpi", 300))
                list_of_figs.append(fig_)
                list_of_paths.append(filepath_) 
 
        if len(list_of_figs)==0 and len(list_of_paths) == 0:
            list_of_figs, list_of_paths = None, None
        
        return list_of_figs, list_of_paths


    def plot_spectrum(
        self, target_time: datetime | np.datetime64, target_range: float, **kwargs
    ) -> tuple[Figure, Path | None]:
        """
        Plots the spectrum for the given target time and range.
        Parameters:
        -----------
        target_time : datetime | np.datetime64
            The target time for which the spectrum is to be plotted.
        target_range : float
            The target range for which the spectrum is to be plotted.
        **kwargs : dict
            Additional keyword arguments to customize the plot. Possible keys include:
            - output_dir : Path, optional
                Directory where the plot will be saved if 'savefig' is True.
            - savefig : bool, optional
                If True, the plot will be saved to a file.
            - ka_color : str, optional
                Color for the Ka-band plot. Default is 'red'.
            - w_color : str, optional
                Color for the W-band plot. Default is 'blue'.
            - dpi : int, optional
                Dots per inch for the saved figure. Default is 300.
        Returns:
        --------
        tuple[Figure, Path | None]
            A tuple containing the figure object and the file path where the figure is saved (if 'savefig' is True).
        """
        _kwargs = kwargs.copy()
        _kwargs.pop("output_dir")
        _kwargs.pop("savefig")
        _kwargs["color"] = kwargs.get("ka_color", "red")
        fig, _ = self.ka.plot_spectrum(target_time, target_range, **_kwargs)

        _kwargs["fig"] = fig
        _kwargs["color"] = kwargs.get("w_color", "blue")
        fig, _ = self.ww.plot_spectrum(target_time, target_range, **_kwargs)
        ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[2]
        ka_line.set_label(f"Ka-band {ka_line.get_label()}")
        w_line.set_label(f"W-band {w_line.get_label()}")

        ax = fig.get_axes()[0]
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"{target_time:%Y%m%dT%H%M%S}_{target_range:.0f}_spectrum.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))

        return fig, filepath



from pathlib import Path
from datetime import datetime

import requests
import numpy as np
import pandas as pd
import xarray as xr
from netCDF4 import Dataset  # type: ignore
from scipy.interpolate import interp1d

from gfatpy.utils.utils import parse_datetime

API_URL = "https://cloudnet.fmi.fi/api"



def get_ecmwf_day(
    target_datetime: datetime | str, times: np.ndarray, heights: np.ndarray, base_query_params = {"site": "granada"}
) -> xr.Dataset:
    """Fetches the ECMWF API to retrieve temperature and pressure data

    Args:

        - target_datetime (dt.datetime | str): A data, hour doesn't affect


    Returns:

        - xr.Dataset: x array dataset with variables: ["pressure", "temperature", "height", "time"]

    Raises:

        - FileNotFoundError: There is no available data in the ECMWF data
    """
    _date = parse_datetime(target_datetime)

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**base_query_params, "date": _date.date().isoformat()},
    )

    if not response_filename.ok:
        raise RuntimeError(
            f"Request returned code {response_filename.status_code}: {response_filename.text}"
        )

    response_filename = response_filename.json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset("ecmwf.nc", memory=file.content)
    dataset = xr.open_dataset(xr.backends.NetCDF4DataStore(nc), decode_timedelta=True)  # type: ignore

    if heights is None:
        heights = np.arange(2666) * 7.5

    d_height = np.tile(heights, (25, 1))
    pressure = np.empty_like(d_height)
    temperature = np.empty_like(d_height)

    hours = np.arange(25)

    for hour in hours:
        pressure[hour, :] = np.interp(
            heights, nc["height"][hour, :], nc["pressure"][hour, :]
        )
        temperature[hour, :] = np.interp(
            heights, nc["height"][hour, :], nc["temperature"][hour, :]
        )

    pressure = np.apply_along_axis(
        lambda row: np.interp(
            times.astype("float64"), dataset.time.values.astype("float64"), row
        ).T,
        0,
        pressure,
    )

    temperature = np.apply_along_axis(
        lambda row: np.interp(
            times.astype("float64"), dataset.time.values.astype("float64"), row
        ).T,
        0,
        temperature,
    )

    return xr.Dataset(
        {
            "temperature": (["time", "range"], temperature),
            "pressure": (["time", "range"], pressure),
        },
        coords={"range": heights, "time": times},
    )  # type: ignore


def get_ecmwf_temperature_pressure(
    date: datetime | str,
    hour: int | None = None,
    heights: np.ndarray | list | None = None,
    base_query_params = {"site": "granada"}
) -> pd.DataFrame:
    """Fetch the ecmwf API and downloads temperature and pressure for a given hour date and hour (optional and default to 12:00). Returns the information in a DataFrame

    Args:

        - date (dt.datetime | str): datetime or str datetime
        - hour (int | None, optional): If need an specific time for the data retieval. Defaults to None.
        - heights (np.ndarray | list | None, optional): A heights profile . Defaults to None.


    Returns:

        - pd.DataFrame: DataFrame with three columns: height, temperature, pressure

    Raises:

        - FileNotFoundError: If ecmwf model doesn't have data available for that date
        - ValueError: Raises in case ranges input array shape is not as expected
    """

    _date = parse_datetime(date)

    if hour is None:
        _date = _date.replace(hour=12)
    else:
        _date = _date.replace(hour=hour)

    hour_int: int = _date.hour

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**base_query_params, "date": _date.date().isoformat()},
    ).json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset("ecmwf.nc", memory=file.content)

    height = nc.variables["height"][hour_int]
    temperature = nc.variables["temperature"][hour_int]
    pressure = nc.variables["pressure"][hour_int]

    if heights is None:
        return pd.DataFrame(
            {"height": height, "temperature": temperature, "pressure": pressure}
        )

    if np.ndim(heights) != 1:
        raise ValueError("heights must be 1-dimensinal")

    # interpolations, case when range doesn't match API heights
    temperature_fn = interp1d(
        height, temperature, bounds_error=False, fill_value="extrapolate"  # type: ignore
    )
    pressure_fn = interp1d(
        height, pressure, bounds_error=False, fill_value="extrapolate"  # type: ignore
    )

    return pd.DataFrame(
        {
            "height": heights,
            "temperature": temperature_fn(heights),
            "pressure": pressure_fn(heights),
        }
    )

def get_data_as_ascii(
    target_datetime: datetime,    
    output_dir: Path | str | None = None,
    base_query_params = {"site": "granada"}
) -> Path:
    """ Fetches temperature and pressure data for a given target datetime,
    saves the data in ASCII format to a file, and returns the path of the saved file.

    Args:

        - target_datetime (datetime.datetime): The target date and time.
        - output_dir (Path | str | None, optional): The directory where the output file    will be saved. If not provided, the current working directory is used.


    Returns:

        - Path: The path of the saved ASCII file.

    Raises:

        - ValueError: If the output_dir is not a valid directory.
        - FileNotFoundError: If no data is found for the given day.
    """
    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(f"output_dir must be a valid directory: {output_dir}")

    hour_int: int = target_datetime.hour

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**base_query_params, "date": target_datetime.date().isoformat()},
    ).json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    # Fetch temperature and pressure data
    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset(output_dir / "ecmwf.nc", memory=file.content)

    height = nc.variables["height"][hour_int]
    temperature = nc.variables["temperature"][hour_int]
    pressure = nc.variables["pressure"][hour_int]

    # Generate ASCII content
    ascii_content = ""
    for idx, h in enumerate(height):
        t = temperature[idx]
        p = pressure[idx]/100.0
        line = f"{h:.2f}\t{t:.2f}\t{p:.2f}\n"
        ascii_content += line

    # Create filename using target_datetime
    current_datetime = target_datetime.strftime("%Y%m%d_%H%M")
    output_file = output_dir / f"{current_datetime}_ecmwf.txt"

    # Save ASCII content to file
    with open(output_file, "w") as file_:
        file_.write("height[m] temperature[K] pressure[hPa]\n")
        file_.write(ascii_content)

    return output_file